import logging
from typing import Optional

from jax._src.prng import PRNGKeyArray

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.map import MAP_NAME
from fortuna.prob_model.posterior.map.map_approximator import MAPPosteriorApproximator
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.map.map_trainer import (
    JittedMAPTrainer,
    MAPTrainer,
    MultiDeviceMAPTrainer,
)
from fortuna.prob_model.posterior.posterior_state_repository import (
    PosteriorStateRepository,
)
from fortuna.typing import Status
from fortuna.utils.builtins import get_dynamic_scale_instance_from_model_dtype
from fortuna.utils.device import select_trainer_given_devices

logger = logging.getLogger(__name__)


class MAPPosterior(Posterior):
    def __init__(
        self,
        joint: Joint,
        posterior_approximator: MAPPosteriorApproximator,
    ):
        """
        Maximum-a-Posteriori (MAP) approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A Joint distribution object.
        posterior_approximator: MAPPosteriorApproximator
            A MAP posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return MAP_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config=None,
        **kwargs,
    ) -> Status:
        super()._checks_on_fit_start(fit_config, map_fit_config)

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            base_trainer_cls=MAPTrainer,
            jitted_trainer_cls=JittedMAPTrainer,
            multi_device_trainer_cls=MultiDeviceMAPTrainer,
            disable_jit=fit_config.processor.disable_jit,
        )

        trainer = trainer_cls(
            predict_fn=self.joint.likelihood.prob_output_layer.predict,
            save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
            eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
            early_stopping_monitor=fit_config.monitor.early_stopping_monitor,
            early_stopping_min_delta=fit_config.monitor.early_stopping_min_delta,
            early_stopping_patience=fit_config.monitor.early_stopping_patience,
            freeze_fun=fit_config.optimizer.freeze_fun,
        )

        if super()._is_state_available_somewhere(fit_config):
            state = self._restore_state_from_somewhere(
                fit_config=fit_config,
                allowed_states=(MAPState,),
            )
        else:
            state = self._init_state(
                data_loader=train_data_loader, fit_config=fit_config
            )

        state = super()._freeze_optimizer_in_state(state, fit_config)
        self._check_state(state)

        logging.info("Run MAP.")
        state, status = trainer.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=self.joint._batched_negative_log_joint_prob,
            training_dataloader=train_data_loader,
            training_dataset_size=train_data_loader.size,
            n_epochs=fit_config.optimizer.n_epochs,
            metrics=fit_config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=val_data_loader.size
            if val_data_loader is not None
            else None,
            verbose=fit_config.monitor.verbose,
            callbacks=fit_config.callbacks,
            max_grad_norm=fit_config.hyperparameters.max_grad_norm,
            gradient_accumulation_steps=fit_config.hyperparameters.gradient_accumulation_steps,
        )
        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        return status

    def sample(self, rng: Optional[PRNGKeyArray] = None, **kwargs) -> JointState:
        state = self.state.get()
        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def _init_state(self, data_loader: DataLoader, fit_config: FitConfig) -> MAPState:
        state = super()._init_joint_state(data_loader=data_loader)

        return MAPState.init(
            params=state.params,
            mutable=state.mutable,
            optimizer=fit_config.optimizer.method,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
            dynamic_scale=get_dynamic_scale_instance_from_model_dtype(
                getattr(self.joint.likelihood.model_manager.model, "dtype")
                if hasattr(self.joint.likelihood.model_manager.model, "dtype")
                else None
            ),
        )



from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
    Union,
)

import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import (
    InputData,
    Mutable,
)


class ModelEditor(nn.Module):
    @nn.compact
    def __call__(
        self,
        apply_fn: Callable[
            [Dict, InputData], Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]
        ],
        model_params: Dict,
        x: Any,
        has_aux: bool,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
        """
        Apply a transformation to the forward pass.

        Parameters
        ----------
        apply_fn: Callable[[Dict, InputData], Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]]
            The model forward pass.
        model_params: Dict
            The model parameters.
        x: Array
            Batch of inputs.
        has_aux: bool
            Whether the forward pass returns auxiliary objects.

        Returns
        -------
        Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]
            Return the transformed outputs, and auxiliary objects if available.
        """
        pass



from fortuna.prob_model.posterior.sgmcmc.sgmcmc_sampling_callback import (
    SGMCMCSamplingCallback,
)
from fortuna.training.callback import Callback
from fortuna.training.train_state import TrainState
from fortuna.training.train_state_repository import TrainStateRepository
from fortuna.training.trainer import TrainerABC


class SGHMCSamplingCallback(SGMCMCSamplingCallback):
    def __init__(
        self,
        n_epochs: int,
        n_training_steps: int,
        n_samples: int,
        n_thinning: int,
        burnin_length: int,
        trainer: TrainerABC,
        state_repository: TrainStateRepository,
        keep_top_n_checkpoints: int,
    ):
        """
        Stochastic Gradient Hamiltonian Monte Carlo (SGHMC) callback that collects samples
        after the initial burn-in phase.

        Parameters
        ----------
        n_epochs: int
            The number of epochs.
        n_training_steps: int
            The number of steps per epoch.
        n_samples: int
            The desired number of the posterior samples.
        n_thinning: int
            Keep only each `n_thinning` sample during the sampling phase.
        burnin_length: int
            Length of the initial burn-in phase, in steps.
        trainer: TrainerABC
            An instance of the trainer class.
        state_repository: TrainStateRepository
            An instance of the state repository.
        keep_top_n_checkpoints: int
            Number of past checkpoint files to keep.
        """
        super().__init__(
            trainer=trainer,
            state_repository=state_repository,
            keep_top_n_checkpoints=keep_top_n_checkpoints,
        )

        self._do_sample = (
            lambda current_step, samples_count: samples_count < n_samples
            and current_step > burnin_length
            and (current_step - burnin_length) % n_thinning == 0
        )

        total_samples = sum(
            self._do_sample(step, 0)
            for step in range(1, n_epochs * n_training_steps + 1)
        )
        if total_samples < n_samples:
            raise ValueError(
                f"The number of desired samples `n_samples` is {n_samples}. However, only "
                f"{total_samples} samples will be collected. Consider adjusting the burnin "
                "length, number of epochs, or the thinning parameter."
            )



import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array


class ClassificationTemperatureScaler(nn.Module):
    r"""
    Classification temperature scaling. It scales the logits with a scalar temperature parameters. Let :math:`o` be
    output logits and :math:`\phi` be a scalar parameter. Then the scaling can be seen as
    :math:`g(\phi, o) = \exp(-\phi) o`.
    """

    @nn.compact
    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        log_temp = self.param("log_temp", nn.initializers.zeros, (1,))
        return x * jnp.exp(-log_temp)



from typing import Optional

from jax import random
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp

from fortuna.prob_model.prior.base import Prior
from fortuna.typing import Params


class IsotropicGaussianPrior(Prior):
    def __init__(self, log_var: Optional[float] = 0.0):
        """
        A diagonal Gaussian prior class.

        Parameters
        ----------
        log_var : Optional[float]
            Prior log-variance value. The covariance matrix of the prior distribution is given by a diagonal matrix
            with this parameter on every entry of the diagonal.
        """
        super().__init__()
        self.log_var = log_var
        self.prec = jnp.exp(-self.log_var)
        self.std = jnp.exp(0.5 * self.log_var)
        self.log2pi = jnp.log(2 * jnp.pi)

    def log_joint_prob(self, params: Params) -> float:
        rav = ravel_pytree(params)[0]
        n = len(rav)
        return -0.5 * (self.prec * jnp.sum(rav**2) + n * (self.log2pi + self.log_var))

    def sample(self, params_like: Params, rng: Optional[PRNGKeyArray] = None) -> Params:
        dummy_rav, unravel = ravel_pytree(params_like)
        n = len(dummy_rav)
        if rng is None:
            rng = self.rng.get()
        rav_samples = self.std * random.normal(rng, shape=(n,))
        return unravel(rav_samples)


class DiagonalGaussianPrior(Prior):
    def __init__(self, log_var: jnp.ndarray):
        """
        A diagonal Gaussian prior class.

        Parameters
        ----------
        log_var : jnp.ndarray
            Prior log-variance vector corresponding to the logarithm of the diagonal of the prior covariance matrix.
        """
        super().__init__()
        self.log_var = log_var
        self.log2pi = jnp.log(2 * jnp.pi)

    def log_joint_prob(self, params: Params) -> float:
        rav = ravel_pytree(params)[0]
        return -0.5 * jnp.sum(
            jnp.exp(-self.log_var) * rav**2 + self.log2pi + self.log_var
        )

    def sample(self, params_like: Params, rng: Optional[PRNGKeyArray] = None) -> Params:
        dummy_rav, unravel = ravel_pytree(params_like)
        n = len(dummy_rav)
        if rng is None:
            rng = self.rng.get()
        rav_samples = jnp.exp(0.5 * self.log_var) * random.normal(rng, shape=(n,))
        return unravel(rav_samples)



from typing import (
    Callable,
    Optional,
    Tuple,
    Union,
)

import jax.numpy as jnp

from fortuna.typing import Array


class Monitor:
    def __init__(
        self,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Union[float, Array]], ...]
        ] = None,
        uncertainty_fn: Optional[
            Callable[[jnp.ndarray, jnp.ndarray, Array], jnp.ndarray]
        ] = None,
        early_stopping_patience: int = 0,
        early_stopping_monitor: str = "val_loss",
        early_stopping_min_delta: float = 0.0,
        eval_every_n_epochs: int = 1,
        disable_calibration_metrics_computation: bool = False,
        verbose: bool = True,
    ):
        """
        An object to configure the monitoring of the calibration process.

        Parameters
        ----------
        metrics: Optional[Callable[[jnp.ndarray, jnp.ndarray, Array], Union[float, Array]]]
            Metrics to monitor during calibration. This must take three arguments: predictions, uncertainty estimates
            and target variables. In classification, :func:`.expected_calibration_error` is an example of valid metric.
        uncertainty_fn: Optional[Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Union[float, Array]], ...]]
            A function that maps (calibrated) outputs into uncertainty estimates. These will be used in `metrics`.
            In classification,
            the default is :func:`~fortuna.prob_output_layer.classification.ProbOutputClassifier.mean`.
            In regression,
            the default is :func:`~fortuna.prob_output_layer.regression.ProbOutputRegressor.variance`.
        early_stopping_patience: int
            Number of consecutive epochs without an improvement in the performance on the validation set before stopping
             the calibration.
        early_stopping_monitor: str
            Validation metric to be monitored for early stopping.
        early_stopping_min_delta: float
            Minimum change between updates to be considered an improvement, i.e., if the absolute change is less than
            `early_stopping_min_delta` then this is not considered an improvement leading to a potential early stop.
        eval_every_n_epochs: int
            Number of calibration epochs between validation. To disable, set
            `eval_every_n_epochs` to None or 0 (i.e., no validation metrics will be computed during calibration).
        disable_calibration_metrics_computation: bool
            if True, during calibration the only metric computed is the objective function.
            Otherwise, all the metrics provided by the user at runtime will be computed for the training step.
        verbose: bool
            Whether to log the calibration progress.
        """
        if metrics is not None:
            if type(metrics) != tuple:
                raise ValueError("`metrics` must be a tuple of callable metrics.")
            for metric in metrics:
                if not callable(metric):
                    raise ValueError(
                        f"All metrics in `metrics` must be callable objects, but {metric} is not."
                    )
        if uncertainty_fn is not None and not callable(uncertainty_fn):
            raise ValueError(f"`uncertainty_fn` must be a a callable function.")

        self.metrics = metrics
        self.uncertainty_fn = uncertainty_fn
        self.early_stopping_patience = early_stopping_patience
        self.early_stopping_monitor = early_stopping_monitor
        self.early_stopping_min_delta = early_stopping_min_delta
        self.eval_every_n_epochs = eval_every_n_epochs
        self.disable_calibration_metrics_computation = (
            disable_calibration_metrics_computation
        )
        self.verbose = verbose



from typing import Any

import flax.linen as nn
import jax.numpy as jnp


class CNN(nn.Module):
    output_dim: int
    dropout_rate: float
    dtype: Any = jnp.float32
    """
    A CNN model.

    :param output_dim: int
        Output dimension.
    :param dropout_rate: Optional[float]
        Dropout rate.
    :param dtype: Any
        Data type. Default: `float32`.
    """

    def setup(self):
        self.hidden_layers = CNNHiddenLayers(
            dropout_rate=self.dropout_rate, dtype=self.dtype
        )
        self.last_layer = CNNLastLayer(output_dim=self.output_dim, dtype=self.dtype)

    def __call__(self, x: jnp.ndarray, train: bool = True) -> jnp.ndarray:
        x = self.hidden_layers(x, train)
        x = self.last_layer(x, train)
        return x


class CNNHiddenLayers(nn.Module):
    dropout_rate: float
    dtype: Any = jnp.float32
    """
    Hidden layers of a CNN model.

    :param dropout_rate: float
        Dropout rate.
    :param dtype: Any
        Data type. Default: `float32`.
    """

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool = True):
        x = nn.Conv(
            features=32,
            kernel_size=(3, 3),
            strides=(1, 1),
            padding="valid",
            dtype=self.dtype,
        )(x)
        x = nn.relu(x)
        x = nn.max_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = nn.Dropout(self.dropout_rate)(x, deterministic=not train)
        x = x.reshape((x.shape[0], -1))
        return x


class CNNLastLayer(nn.Module):
    output_dim: int
    dtype: Any = jnp.float32
    """
    Last layer of a CNN model.

    :param output_dim: int
        Output dimension.
    :param dtype: Any
        Data type. Default: `float32`.
    """

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool = True):
        x = nn.Dense(features=128, dtype=self.dtype)(x)
        x = nn.Dense(features=self.output_dim, dtype=self.dtype)(x)
        return x



from typing import (
    Optional,
    Union,
)

from jax import vmap
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
import numpy as np

from fortuna.data.loader import InputsLoader
from fortuna.likelihood.base import Likelihood
from fortuna.model.model_manager.regression import RegressionModelManager
from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.prob_output_layer.regression import RegressionProbOutputLayer
from fortuna.typing import (
    Array,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
)


class RegressionLikelihood(Likelihood):
    def __init__(
        self,
        model_manager: RegressionModelManager,
        prob_output_layer: RegressionProbOutputLayer,
        output_calib_manager: Optional[OutputCalibManager] = None,
    ):
        """
        A regression likelihood function class. In this class, the likelihood function is additionally assumed to
        be a probability density function, i.e. positive and integrating to 1. The likelihood is formed by three
        objects applied in sequence: the model manager, the output calibrator and the probabilistic output layer. The
        model manager maps parameters and inputs to outputs. The output calibration takes outputs and returns some
        calibrated version of them. The probabilistic output layer describes the probability distribution of the
        calibrated outputs.

        Parameters
        ----------
        model_manager : ModelManager
            An model manager. This objects orchestrates the evaluation of the models.
        prob_output_layer : ProbOutputLayer
            A probabilistic output layer object. This object characterizes the probability distribution of the target
            variable given the calibrated outputs.
        output_calib_manager : Optional[OutputCalibManager]
            An output calibration manager object. It transforms outputs of the model manager into some
            calibrated version of them.
        """
        super().__init__(
            model_manager, prob_output_layer, output_calib_manager=output_calib_manager
        )

    def _batched_mean(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        outputs = super()._get_batched_calibrated_outputs(
            params, inputs, mutable, calib_params, calib_mutable, **kwargs
        )
        return outputs[:, : outputs.shape[1] // 2]

    def _batched_mode(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        return self._batched_mean(
            params,
            inputs,
            mutable,
            calib_params=calib_params,
            calib_mutable=calib_mutable,
            **kwargs,
        )

    def _batched_variance(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        **kwargs,
    ) -> jnp.ndarray:
        outputs = super()._get_batched_calibrated_outputs(
            params, inputs, mutable, calib_params, calib_mutable, **kwargs
        )
        return jnp.exp(outputs[:, outputs.shape[1] // 2 :])

    def entropy(
        self,
        params: Params,
        inputs_loader: InputsLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        n_target_samples: Optional[int] = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        samples, aux = self.sample(
            n_target_samples,
            params,
            inputs_loader,
            mutable,
            calib_params=calib_params,
            calib_mutable=calib_mutable,
            return_aux=["outputs"],
            rng=rng,
            distribute=distribute,
        )
        outputs = aux["outputs"]

        @vmap
        def _log_lik_fun(sample: jnp.ndarray):
            return self.prob_output_layer.log_prob(outputs, sample, **kwargs)

        return -jnp.mean(_log_lik_fun(samples), 0)

    def quantile(
        self,
        q: Union[float, jnp.ndarray, np.ndarray],
        params: Optional[Params] = None,
        inputs_loader: Optional[InputsLoader] = None,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        n_target_samples: Optional[int] = 30,
        target_samples: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> Union[float, jnp.ndarray]:
        """
        Estimate the `q`-th quantiles of the likelihood function.

        Parameters
        ----------
        q: Union[float, jnp.ndarray, np.ndarray]
            Quantile or sequence of quantiles to compute, which must be between 0 and 1 inclusive.
        params : Params
            The random parameters of the probabilistic model.
        inputs_loader : InputsLoader
            A loader of input data points.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        n_target_samples : int
            Number of target samples to sample for each input data point.
        target_samples: Optional[jnp.ndarray] = None
            Samples of the target variable for each input, used to estimate the quantiles.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            Quantile estimate for each quantile and each input. If multiple quantiles `q` are given, the result's
            first axis is over different quantiles.
        """
        if target_samples is None:
            if params is None or inputs_loader is None:
                raise ValueError(
                    "if `samples` is not passed, then `params` and `inputs_loader` must be passed."
                )
            target_samples = self.sample(
                n_target_samples,
                params,
                inputs_loader,
                mutable,
                calib_params=calib_params,
                calib_mutable=calib_mutable,
                rng=rng,
                distribute=distribute,
                **kwargs,
            )
        return jnp.quantile(target_samples, q, axis=0)



from typing import (
    Dict,
    Optional,
)

from fortuna.prob_model.posterior.posterior_mixin import WithPosteriorCheckpointingMixin
from fortuna.training.train_state_repository import TrainStateRepository
from fortuna.typing import Path


class PosteriorStateRepository(WithPosteriorCheckpointingMixin, TrainStateRepository):
    def extract_calib_keys(
        self,
        checkpoint_path: Optional[Path] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Dict:
        return super().extract(
            ["calib_params", "calib_mutable"], checkpoint_path, prefix, **kwargs
        )



import abc

from fortuna.utils.builtins import HashableMixin


class Distribution(HashableMixin, abc.ABC):
    @abc.abstractmethod
    def sample(self, *args, **kwargs):
        pass

    def log_joint_prob(self, *args, **kwargs):
        pass



from __future__ import annotations

from typing import (
    Dict,
    Optional,
    Union,
)

from flax.core import FrozenDict

from fortuna.typing import (
    Mutable,
    Params,
)


class ModelManagerState:
    params: Params
    mutable: Optional[Mutable] = None

    def __init__(self, params: Params, mutable: Optional[Mutable] = None):
        """
        An model manager state class.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        """
        self.params = params
        self.mutable = mutable

    @classmethod
    def init_from_dict(cls, d: Union[Dict, FrozenDict]) -> ModelManagerState:
        """
        Initialize the model manager state from a dictionary. This dictionary should be like the output of
        :func:`~fortuna.model.model_manager.base.ModelManager.init`.

        Parameters
        ----------
        d : Union[Dict, FrozenDict]
            A dictionary like the output of :func:`~fortuna.model.model_manager.base.ModelManager.init`.

        Returns
        -------
        ModelManagerState
            An model manager state.
        """
        params = FrozenDict(
            {k: FrozenDict({"params": v["params"]}) for k, v in d.items()}
        )
        mutable = FrozenDict(
            {
                k: FrozenDict({_k: _v for _k, _v in v.items() if _k != "params"})
                for k, v in d.items()
            }
        )
        flag = 0
        for k, v in mutable.items():
            if len(v) > 0:
                flag += 1
        if flag == 0:
            mutable = None
        return cls(params=params, mutable=mutable)



from __future__ import annotations

import jax.numpy as jnp

from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.utils.strings import convert_string_to_jnp_array


class MAPState(PosteriorState):
    """
    Attributes
    ----------
    encoded_name: jnp.ndarray
        MAP state name encoded as an array.
    """

    encoded_name: jnp.ndarray = convert_string_to_jnp_array("MAPState")



from typing import (
    Callable,
    List,
    Optional,
    Tuple,
    Union,
)

import matplotlib
from matplotlib.collections import Collection
from matplotlib.patches import (
    Circle,
    RegularPolygon,
)
from matplotlib.path import Path
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
import matplotlib.pyplot as plt
from matplotlib.spines import Spine
from matplotlib.transforms import Affine2D
import numpy as np

from fortuna.data.loader import InputsLoader
from fortuna.prob_model.regression import ProbRegressor
from fortuna.typing import (
    Array,
    Batch,
)


def plot_2d_classification_predictions_and_uncertainty(
    inputs: Optional[np.ndarray] = None,
    preds: Optional[np.ndarray] = None,
    grid_uncertainty: Optional[np.ndarray] = None,
    uncertainty_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    grid: Optional[np.ndarray] = None,
    xx1: Optional[np.ndarray] = None,
    xx2: Optional[np.ndarray] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    figsize: tuple = (5, 5),
    title: Optional[str] = None,
    uncertainty_cmap: str = "Blues",
    preds_color: Tuple[str, str] = ("C2", "C3"),
    base_inputs_color: str = "black",
    marker_size: int = 10,
    legend: bool = True,
    uncertainty_label: str = "uncertainty",
    fontsize: int = 12,
    colorbar: bool = False,
    remove_ticks: bool = False,
    show: bool = False,
    return_pcolor: bool = False,
    **save_options,
) -> Optional[Collection]:
    """
    Plot inputs, predictions and uncertainty.

    :param inputs: Optional[np.ndarray]
        Input data. This must be a two-dimensional array, with different inputs over the first axis. The number of
        dimensions over the second axis must be 2.
    :param preds: Optional[np.ndarray]
        Predictions. This must be a one-dimensional array of zeros and ones, with different inputs over the first
        axis.
    :param grid_uncertainty: Optional[np.ndarray]
        Scalar uncertainty value for each element of the grid. This must have the same dimensionality as `grid`.
    :param uncertainty_fn: Optional[Callable[[np.ndarray], np.ndarray]]
        A function taking as argument a grid of inputs, and returning a scalar uncertainty value for each element
        in the grid.
    :param grid: Optional[np.ndarray]
        The grid to display the plot over.
    :param xx1: Optional[np.ndarray]
        Mesh over the first axis.
    :param xx2: Optional[np.ndarray]
        Mesh over the second axis.
    :param ax: Optional[matplotlib.axes.Axes]
        Axis matplotlib object.
    :param figsize: tuple
        Figure size.
    :param title: Optional[str]
        Figure title.
    :param uncertainty_cmap: str
        Uncertainty color map.
    :param preds_color: Tuple[str, str]
        Color of predictions. The first element of the tuple is associated to the prediction 1; the second to 0.
    :param base_inputs_color: str
        Color of inputs if `preds` is not available.
    :param marker_size: int
        Size of scattered inputs.
    :param legend: bool
        Whether to include a legend.
    :param uncertainty_label: str
        Uncertainty label in the legend.
    :param fontsize: int
        Font size in the legend.
    :param colorbar: bool
        Whether to include a color bar.
    :param remove_ticks: bool
        Whether to remove ticks from axes.
    :param save_options: dict
        Options to save the file with `matplotlib.pyplot.savefig`. If no option is given, the file is not saved.
    :param show: bool
        Whether to show the plot.
    :param return_pcolor: bool
        Whether to return the output of `matplotlib.pyplot.pcolor`.

    :return Optional[Collection]
        If `return_pcolor`, it returns the output of `matplotlib.pyplot.pcolor`.
    """
    if (xx1 is None or xx2 is None) and (inputs is None):
        raise ValueError(
            """If `inputs` is not passed, please pass either both `xx1` and `xx2`, or `grid`.
        """
        )
    if grid_uncertainty is None and uncertainty_fn is None:
        raise ValueError(
            """ One between `grid_uncertainty` and `uncertainty_fn` must be passed. If both are passed,
        `uncertainty_fn` is ignored."""
        )
    if grid is None and uncertainty_fn is None:
        raise ValueError(
            """If `grid` is not passed, then `uncertainty_fn` must be passed."""
        )
    if grid is not None and (uncertainty_fn is None and grid_uncertainty is None):
        raise ValueError(
            """If `grid` is not passed, then at least one of `uncertainty_fn` and `grid_uncertainty`
        must be passed."""
        )
    if grid_uncertainty is not None and grid is None:
        raise ValueError(
            """If `grid_uncertainty` is passed, then `grid` must also be passed."""
        )
    if grid is None:
        if xx1 is None or xx2 is None:
            mins, maxs = np.min(inputs, 0), np.max(inputs, 0)
            if xx1 is None:
                xx1 = np.linspace(
                    mins[0] - 1.25 * (maxs[0] - mins[0]),
                    maxs[0] + 1.25 * (maxs[0] - mins[0]),
                    100 if xx2 is None else len(xx2),
                )
            if xx2 is None:
                xx2 = np.linspace(
                    mins[1] - 1.25 * (maxs[1] - mins[1]),
                    maxs[1] + 1.25 * (maxs[1] - mins[1]),
                    100 if xx1 is None else len(xx1),
                )
        grid = np.array([[_xx1, _xx2] for _xx1 in xx1 for _xx2 in xx2])
    if grid_uncertainty is None:
        grid_uncertainty = uncertainty_fn(grid)
    size1 = len(np.unique(grid[:, 0])) if xx1 is None else len(xx1)
    size2 = len(np.unique(grid[:, 1])) if xx2 is None else len(xx2)
    grid_uncertainty = grid_uncertainty.reshape(size1, size2)
    grid = grid.reshape(size1, size2, 2)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    if title:
        ax.set_title(title, fontsize=fontsize + 2)
    im = ax.pcolor(
        grid[:, :, 0],
        grid[:, :, 1],
        grid_uncertainty,
        cmap=uncertainty_cmap,
        label=uncertainty_label,
    )
    if inputs is not None:
        ax.scatter(
            inputs[:, 0],
            inputs[:, 1],
            s=marker_size,
            c=[preds_color[0] if i == 1 else preds_color[1] for i in preds]
            if preds is not None
            else base_inputs_color,
        )
    if colorbar:
        plt.colorbar(im, ax=ax.ravel().tolist() if hasattr(ax, "ravel") else ax)
    if legend:
        ax.legend(fontsize=fontsize)
    if remove_ticks:
        ax.xticks([], [])
        ax.yticks([], [])
    if len(save_options) > 0:
        plt.savefig(**save_options)
    if show:
        plt.show()

    if return_pcolor:
        return im


def plot_reliability_diagram(
    accs: Union[Array, List[Array]],
    confs: Union[Array, List[Array]],
    ax: Optional[matplotlib.axes.Axes] = None,
    labels: Optional[List[str]] = None,
    figsize: tuple = (6, 4),
    legend_loc: Optional[str] = None,
    fontsize: int = 10,
    title: Optional[str] = None,
    ylim: Optional[Tuple[float, float]] = None,
    show: bool = False,
    **save_options,
) -> None:
    """
    Plot a reliability diagram.

    Parameters
    ----------
    accs: Union[Array, List[Array, ...]]
        An accuracy score for each bin. A list of multiple accuracy scores is also accepted.
    confs: Union[Array, List[Array, ...]]
        A confidence score for each bin. A list of multiple confidence scores is also accepted.
    labels: Optional[List[str, ...]]
        Labels used in the legend presumably describing the method used to generate accuracy and confidence scores. A
        list of labels, corresponding to lists of accuracies and confidences, is also accepted.
    ax: Optional[matplotlib.axes.Axes]
        Axis matplotlib object.
    figsize: Optional[tuple]
        Figure size.
    legend_loc: Optional[str]
        Legend location, as in :code:`matplotlib`.
        See `here <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html>`__
    fontsize: Optional[int]
        Font size.
    title: Optional[str] = None
        Plot title.
    ylim: Optional[Tuple[float, float]]
        Bottom and top limits on the y-axis.
    show: bool
        Whether to show the plot.
    save_options: dict
        Options to save the file with `matplotlib.pyplot.savefig`. If no option is given, the file is not saved.
    """
    if figsize is None:
        figsize = (6, 4)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    if type(confs) != list:
        confs = [confs]
    if type(accs) != list:
        accs = [accs]
    if labels:
        if type(labels) != list:
            labels = [labels]
    if len(accs) != len(confs):
        raise ValueError("`accs` and `confs` must contain the same number of entries.")
    if labels:
        if len(accs) != len(labels):
            raise ValueError(
                "`accs` and `labels` must contain the same number of entries."
            )
    ax.grid()
    ax.plot([0, 1], [0, 0], color="grey", linestyle="--", alpha=0.3)
    for i, (a, c) in enumerate(zip(accs, confs)):
        ax.plot(
            c, c - a, marker=".", linestyle="-", label=labels[i] if labels else None
        )
    if labels:
        ax.legend(fontsize=fontsize, loc=legend_loc if legend_loc else None)
    ax.set_xlabel("confidence", fontsize=fontsize)
    ax.set_ylabel("confidence - accuracy", fontsize=fontsize)
    if title:
        ax.set_title(title, fontsize=fontsize + 2)
    if ylim:
        ax.set_ylim(ylim)
    if len(save_options) > 0:
        plt.savefig(**save_options)
    if show:
        plt.show()
    if ax:
        return ax


def plot_1d_regression_preds_and_std(
    prob_model: ProbRegressor,
    mesh: np.ndarray,
    data: Optional[Batch] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    n_std: float = 2.0,
    title: Optional[str] = None,
    figsize: Tuple = (5, 2),
    fontsize: int = 12,
    marker_size_data: int = 1,
    legend: bool = True,
    show: bool = False,
    **saving_options,
) -> Optional[Collection]:
    mesh_loader = InputsLoader.from_array_inputs(mesh)
    if not ax:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    if title:
        ax.set_title(title, fontsize=fontsize + 2)

    mean = prob_model.predictive.mean(mesh_loader)
    std = prob_model.predictive.std(mesh_loader)
    if data:
        ax.scatter(*data, s=marker_size_data, color="C0", label="data")
    ax.plot(mesh, mean, color="C1", label="mean")
    ax.fill_between(
        mesh,
        (mean - n_std * std).squeeze(1),
        (mean + n_std * std).squeeze(1),
        alpha=0.3,
        color="C0",
        label=f"+/- {n_std} std",
    )
    if legend:
        ax.legend(fontsize=fontsize - 2)
    plt.tight_layout()
    if show:
        plt.show()
    if len(saving_options) > 0:
        plt.savefig(**saving_options)
    return ax


def radar_factory(num_vars, frame="circle"):
    """
    Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle', 'polygon'}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False)

    class RadarTransform(PolarAxes.PolarTransform):
        def transform_path_non_affine(self, path):
            # Paths with non-unit interpolation steps correspond to gridlines,
            # in which case we force interpolation (to defeat PolarTransform's
            # autoconversion to circular arcs).
            if path._interpolation_steps > 1:
                path = path.interpolated(num_vars)
            return Path(self.transform(path.vertices), path.codes)

    class RadarAxes(PolarAxes):
        name = "radar"
        PolarTransform = RadarTransform

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location("N")

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.append(x, x[0])
                y = np.append(y, y[0])
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
            # in axes coordinates.
            if frame == "circle":
                return Circle((0.5, 0.5), 0.5)
            elif frame == "polygon":
                return RegularPolygon((0.5, 0.5), num_vars, radius=0.5, edgecolor="k")
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

        def _gen_axes_spines(self):
            if frame == "circle":
                return super()._gen_axes_spines()
            elif frame == "polygon":
                # spine_type must be 'left'/'right'/'top'/'bottom'/'circle'.
                spine = Spine(
                    axes=self,
                    spine_type="circle",
                    path=Path.unit_regular_polygon(num_vars),
                )
                # unit_regular_polygon gives a polygon of radius 1 centered at
                # (0, 0) but we want a polygon of radius 0.5 centered at (0.5,
                # 0.5) in axes coordinates.
                spine.set_transform(
                    Affine2D().scale(0.5).translate(0.5, 0.5) + self.transAxes
                )
                return {"polar": spine}
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)
    return theta



from flax.core import FrozenDict
import jax.numpy as jnp

from fortuna.training.callback import Callback
from fortuna.training.train_state import TrainState
from fortuna.utils.nested_dicts import (
    find_one_path_to_key,
    nested_get,
    nested_update,
)


class ResetCovarianceCallback(Callback):
    """
    Reset, at the beginning of each epoch, the covariance matrix estimated while training an SNGP model.
    """

    def __init__(self, precision_matrix_key_name: str, ridge_penalty: float):
        self.precision_matrix_key_name = precision_matrix_key_name
        self.ridge_penalty = ridge_penalty

    def training_epoch_start(self, state: TrainState) -> TrainState:
        key_paths = find_one_path_to_key(state.mutable, self.precision_matrix_key_name)
        precision_matrix = nested_get(state.mutable, key_paths)
        if precision_matrix.ndim == 2:
            n, _ = precision_matrix.shape  # rows, cols
            init_precision_matrix = (
                jnp.eye(n, dtype=precision_matrix.dtype) * self.ridge_penalty
            )
        elif precision_matrix.ndim == 3:
            d, n, _ = precision_matrix.shape  # num_devices, rows, cols
            init_precision_matrix = (
                jnp.eye(n, dtype=precision_matrix.dtype) * self.ridge_penalty
            )
            init_precision_matrix = jnp.broadcast_to(init_precision_matrix, (d, n, n))

        partially_updated_mutables = init_precision_matrix
        for key in reversed(key_paths):
            partially_updated_mutables = {key: partially_updated_mutables}
        mutables = nested_update(state.mutable.unfreeze(), partially_updated_mutables)
        mutables = FrozenDict(mutables)
        return state.replace(mutable=mutables)



import abc
import logging
from typing import (
    Callable,
    Dict,
    Optional,
)

import jax
import jax.numpy as jnp

from fortuna.data.loader import DataLoader
from fortuna.output_calib_model.state import OutputCalibState
from fortuna.prob_model.calib_config.base import CalibConfig
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.prob_model_calibrator import (
    JittedProbModelOutputCalibrator,
    MultiDeviceProbModelOutputCalibrator,
    ProbModelOutputCalibrator,
)
from fortuna.typing import (
    Array,
    Path,
    Status,
)
from fortuna.utils.data import check_data_loader_is_not_random
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.random import RandomNumberGenerator


class ProbModel(abc.ABC):
    """
    Abstract probabilistic model class.
    """

    def __init__(self, seed: int = 0):
        self.rng = RandomNumberGenerator(seed=seed)
        self.__set_rng()

    def __set_rng(self):
        self.model_manager.rng = self.rng
        self.output_calib_manager.rng = self.rng
        self.prob_output_layer.rng = self.rng
        self.prior.rng = self.rng
        self.likelihood.rng = self.rng
        self.joint.rng = self.rng
        self.posterior.rng = self.rng
        self.predictive.rng = self.rng

    def train(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        calib_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        calib_config: CalibConfig = CalibConfig(),
        map_fit_config: Optional[FitConfig] = None,
    ) -> Dict[str, Status]:
        """
        Train the probabilistic model. This involves fitting the posterior distribution and calibrating the
        probabilistic model. Calibration is performed only if (1) `calib_data_loader` is passed and (2) the
        probabilistic model contains any calibrator.

        Parameters
        ----------
        train_data_loader : DataLoader
            A training data loader.
        val_data_loader : DataLoader
            A validation data loader. This is used to validate both posterior fitting and calibration.
        calib_data_loader : DataLoader
            A calibration data loader. If this is not passed, no calibration is performed.
        fit_config : FitConfig
            An object to configure the posterior distribution fitting.
        calib_config : CalibConfig
            An object to configure the calibration.
        map_fit_config : Optional[FitConfig] = None
            An object to configure a preliminary posterior distribution fitting via the Maximum-A-Posteriori (MAP)
            method.
            The fit methods of several supported posterior approximations, like the ones of
            :class:`~fortuna.prob_model.posterior.swag.swag_posterior.SWAGPosterior` and
            :class:`~fortuna.prob_model.posterior.laplace.laplace_posterior.LaplacePosterior`, start from a preliminary
            run of MAP, which can be configured via this object. If the method does not start from MAP, this argument is
            ignored.

        Returns
        -------
        Dict[str, Status]
            Status objects for both posterior fitting and calibration.

        """
        logging.info("Fit the posterior distribution...")
        fit_status = self.posterior.fit(
            train_data_loader=train_data_loader,
            val_data_loader=val_data_loader,
            fit_config=fit_config,
            map_fit_config=map_fit_config,
        )

        calib_status = None
        if calib_data_loader:
            calib_status = self.calibrate(
                calib_data_loader=calib_data_loader,
                val_data_loader=val_data_loader,
                calib_config=calib_config,
            )
            logging.info("Calibration completed.")
        return dict(fit_status=fit_status, calib_status=calib_status)

    def _calibrate(
        self,
        calib_data_loader: DataLoader,
        uncertainty_fn: Callable[[jnp.ndarray, jnp.ndarray, Array], jnp.ndarray],
        val_data_loader: Optional[DataLoader] = None,
        calib_config: CalibConfig = CalibConfig(),
    ) -> Status:
        check_data_loader_is_not_random(calib_data_loader)
        if val_data_loader is not None:
            check_data_loader_is_not_random(val_data_loader)
        if (
            self.output_calib_manager is None
            or self.output_calib_manager.output_calibrator is None
        ):
            logging.warning(
                """Nothing to calibrate. No calibrator was passed to the probabilistic model."""
            )
        else:
            if self.posterior.state is None:
                raise ValueError(
                    """Before calibration, you must either train the probabilistic model (see
                        :meth:`~fortuna.prob_model.base.ProbModel.train`),
                        or load a state from an existing checkpoint
                        (see :meth:`~fortuna.prob_model.base.ProbModel.load_state`)."""
                )
            if calib_config.monitor.verbose:
                logging.info(
                    "Pre-compute ensemble of outputs on the calibration data loader."
                )

            distribute = jax.local_devices()[0].platform != "cpu"

            (
                calib_ensemble_outputs_loader,
                calib_size,
            ) = self.predictive._sample_outputs_loader(
                inputs_loader=calib_data_loader.to_inputs_loader(),
                n_output_samples=calib_config.processor.n_posterior_samples,
                return_size=True,
                distribute=distribute,
            )
            if calib_config.monitor.verbose:
                logging.info(
                    "Pre-compute ensemble of outputs on the validation data loader."
                )
            val_ensemble_outputs_loader, val_size = (
                self.predictive._sample_outputs_loader(
                    inputs_loader=val_data_loader.to_inputs_loader(),
                    n_output_samples=calib_config.processor.n_posterior_samples,
                    return_size=True,
                    distribute=distribute,
                )
                if val_data_loader is not None
                else (None, None)
            )

            trainer_cls = select_trainer_given_devices(
                devices=calib_config.processor.devices,
                base_trainer_cls=ProbModelOutputCalibrator,
                jitted_trainer_cls=JittedProbModelOutputCalibrator,
                multi_device_trainer_cls=MultiDeviceProbModelOutputCalibrator,
                disable_jit=calib_config.processor.disable_jit,
            )

            calibrator = trainer_cls(
                calib_outputs_loader=calib_ensemble_outputs_loader,
                val_outputs_loader=val_ensemble_outputs_loader,
                predict_fn=self.prob_output_layer.predict,
                uncertainty_fn=uncertainty_fn,
                save_checkpoint_dir=calib_config.checkpointer.save_checkpoint_dir,
                save_every_n_steps=calib_config.checkpointer.save_every_n_steps,
                keep_top_n_checkpoints=calib_config.checkpointer.keep_top_n_checkpoints,
                disable_training_metrics_computation=calib_config.monitor.disable_calibration_metrics_computation,
                eval_every_n_epochs=calib_config.monitor.eval_every_n_epochs,
                early_stopping_monitor=calib_config.monitor.early_stopping_monitor,
                early_stopping_min_delta=calib_config.monitor.early_stopping_min_delta,
                early_stopping_patience=calib_config.monitor.early_stopping_patience,
            )

            if calib_config.checkpointer.restore_checkpoint_path is None:
                calib_dict = self.posterior.state.extract_calib_keys()

                state = OutputCalibState.init(
                    params=calib_dict["calib_params"],
                    mutable=calib_dict["calib_mutable"],
                    optimizer=calib_config.optimizer.method,
                )
            else:
                state = self.posterior.restore_checkpoint(
                    calib_config.checkpointer.restore_checkpoint_path,
                    optimizer=calib_config.optimizer.method,
                )

            if calib_config.monitor.verbose:
                logging.info("Start calibration.")
            state, status = calibrator.train(
                rng=self.rng.get(),
                state=state,
                loss_fun=self.predictive._batched_negative_log_joint_prob,
                training_data_loader=calib_data_loader,
                training_dataset_size=calib_size,
                n_epochs=calib_config.optimizer.n_epochs,
                metrics=calib_config.monitor.metrics,
                val_data_loader=val_data_loader,
                val_dataset_size=val_size,
                verbose=calib_config.monitor.verbose,
            )

            self.posterior.state.update(
                variables=dict(calib_params=state.params, calib_mutable=state.mutable)
            )

            if (
                calib_config.checkpointer.dump_state
                and calib_config.checkpointer.save_checkpoint_dir is not None
            ):
                if calib_config.monitor.verbose:
                    logging.info("Dump state to disk.")
                self.save_state(
                    checkpoint_path=calib_config.checkpointer.save_checkpoint_dir
                )

            if calib_config.monitor.verbose:
                logging.info("Calibration completed.")

            return status

    def load_state(self, checkpoint_path: Path) -> None:
        """
        Load the state of the posterior distribution from a checkpoint path. The checkpoint must be compatible with the
        probabilistic model.

        Parameters
        ----------
        checkpoint_path : Path
            Path to a checkpoint file or directory to restore.
        """
        return self.posterior.load_state(checkpoint_path)

    def save_state(
        self, checkpoint_path: Path, keep_top_n_checkpoints: int = 1
    ) -> None:
        """
        Save the posterior distribution state as a checkpoint.

        Parameters
        ----------
        checkpoint_path : Path
            Path to file or directory where to save the current state.
        keep_top_n_checkpoints : int
            Number of past checkpoint files to keep.
        """
        return self.posterior.save_state(checkpoint_path, keep_top_n_checkpoints)



from jax import vmap
import jax.numpy as jnp

from fortuna.conformal.classification.base import (
    CVPlusConformalClassifier,
    SplitConformalClassifier,
)
from fortuna.typing import Array


@vmap
def _score_fn(probs: Array, perm: Array, inv_perm: Array, targets: Array):
    return jnp.cumsum(probs[perm])[inv_perm[targets]]


def score_fn(
    probs: Array,
    targets: Array,
):
    perms = jnp.argsort(probs, axis=1)[:, ::-1]
    inv_perms = jnp.argsort(perms, axis=1)
    return _score_fn(probs, perms, inv_perms, targets)


class AdaptivePredictionConformalClassifier(SplitConformalClassifier):
    def score_fn(
        self,
        probs: Array,
        targets: Array,
    ):
        return score_fn(probs=probs, targets=targets)


class CVPlusAdaptivePredictionConformalClassifier(CVPlusConformalClassifier):
    def score_fn(
        self,
        probs: Array,
        targets: Array,
    ):
        return score_fn(probs=probs, targets=targets)



import logging
from typing import Optional

from fortuna.data import DataLoader
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.posterior.map.map_posterior import MAPPosterior
from fortuna.prob_model.posterior.sngp import SNGP_NAME
from fortuna.prob_model.posterior.sngp.sngp_approximator import (
    SNGPPosteriorApproximator,
)
from fortuna.prob_model.posterior.sngp.sngp_callback import ResetCovarianceCallback
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import Status
from fortuna.utils.nested_dicts import find_one_path_to_key

logger = logging.getLogger(__name__)


class SNGPPosterior(MAPPosterior):
    def __init__(
        self,
        joint: Joint,
        posterior_approximator: SNGPPosteriorApproximator,
    ):
        """
        Spectral-normalized Neural Gaussian Process (`SNGP <https://arxiv.org/abs/2006.10108>`_) approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A Joint distribution object.
        posterior_approximator: SNGPPosteriorApproximator
            An SNGP posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        **kwargs,
    ) -> Status:
        # set sngp callback to reset covariance
        callbacks = [
            ResetCovarianceCallback(
                precision_matrix_key_name="precision_matrix",
                ridge_penalty=self.joint.likelihood.model_manager.ridge_penalty,
            )
        ]
        if fit_config.callbacks is None:
            fit_config.callbacks = callbacks
        else:
            fit_config.callbacks = fit_config.callbacks + callbacks
        return super(SNGPPosterior, self).fit(
            train_data_loader, val_data_loader, fit_config, **kwargs
        )

    def __str__(self):
        return SNGP_NAME

    @staticmethod
    def _check_state(state: PosteriorState) -> None:
        path = find_one_path_to_key(state.mutable, "spectral_stats")
        if len(path) == 0:
            raise ValueError(
                f"It looks like your deep feature extractor does not have Spectral Normalization, "
                f"which is required by SNGP. Please include spectral normalization in your model."
                f"Check out `fortuna.model.utils.spectral_norm.WithSpectralNorm` for more details."
            )



from typing import Optional

import jax
from jax import (
    random,
    vmap,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
import jax.scipy as jsp

from fortuna.prob_output_layer.base import ProbOutputLayer
from fortuna.typing import Array


class ClassificationProbOutputLayer(ProbOutputLayer):
    def __init__(self):
        r"""
        Classification probabilistic output layers class. It characterizes the probability distribution of a target
        variable given a calibrated output logits as a Categorical distribution. That is
        :math:`p(y|\omega)=\text{Categorical}(y|p=\text{softmax}(\omega))`, where :math:`y` denotes a target variable
        and :math:`\omega` a calibrated output.
        """
        super().__init__()

    def log_prob(self, outputs: Array, targets: Array, **kwargs) -> Array:
        n_cats = outputs.shape[-1]
        targets = jax.nn.one_hot(targets, n_cats)
        return jnp.sum(targets * outputs, -1) - jsp.special.logsumexp(outputs, -1)

    def predict(self, outputs: Array, **kwargs) -> jnp.ndarray:
        return jnp.argmax(outputs, -1)

    def sample(
        self,
        n_target_samples: int,
        outputs: Array,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> jnp.ndarray:
        probs = (
            jax.nn.softmax(outputs, -1)
            if "probs" not in kwargs or kwargs["probs"] is None
            else kwargs["probs"]
        )
        n_cats = probs.shape[-1]
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, probs.shape[0])
        return vmap(
            lambda key, p: random.choice(key, n_cats, p=p, shape=(n_target_samples,)),
            out_axes=1,
        )(keys, probs)

    def mean(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the mean of the one-hot encoded target variable given the output with respect to the probabilistic
        output layer distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated mean for each output.
        """
        return jax.nn.softmax(outputs, -1)

    def mode(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the mode of the one-hot encoded target variable given the output with respect to the probabilistic
        output layer distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated mode for each output.
        """
        return jnp.argmax(outputs, -1)

    def variance(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the variance of the one-hot encoded target variable given the output with respect to the probabilistic
        output layer distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated variance for each output.
        """
        p = self.mean(outputs)
        return p * (1 - p)

    def std(self, outputs: Array, variances: Optional[Array] = None) -> jnp.ndarray:
        """
        Estimate the standard deviation of the one-hot encoded target variable given the output with respect to the
        probabilistic output layer distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs
        variances: Optional[Array]
            Variance for each output.

        Returns
        -------
        jnp.ndarray
            The estimated standard deviation for each output.
        """
        return super().std(outputs, variances)

    def entropy(self, outputs: Array, **kwargs) -> jnp.ndarray:
        """
        Estimate the entropy of the one-hot encoded target variable given the output with respect to the probabilistic
        output layer distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs

        Returns
        -------
        jnp.ndarray
            The estimated entropy for each output.
        """
        n_classes = outputs.shape[-1]

        @vmap
        def _entropy_term(i: int):
            targets = i * jnp.ones(outputs.shape[0])
            log_liks = self.log_prob(outputs, targets, **kwargs)
            return jnp.exp(log_liks) * log_liks

        return -jnp.sum(_entropy_term(jnp.arange(n_classes)), 0)


class ClassificationMaskedProbOutputLayer(ClassificationProbOutputLayer):
    def log_prob(self, outputs: Array, targets: Array, **kwargs) -> Array:
        n_cats = outputs.shape[-1]
        targets_mask = jnp.where(targets > 0, 1.0, 0.0)
        targets = jax.nn.one_hot(targets, n_cats)
        return (
            jnp.sum(targets * outputs, -1) - jsp.special.logsumexp(outputs, -1)
        ) * targets_mask



from fortuna.conformal.multivalid.iterative.batch_mvp import BatchMVPConformalMethod
from fortuna.conformal.regression.base import ConformalRegressor


class BatchMVPConformalRegressor(BatchMVPConformalMethod, ConformalRegressor):
    def __init__(self, seed: int = 0):
        """
        This class implements a classification version of BatchMVP
        `[Jung et al., 2022] <https://arxiv.org/abs/2209.15145>`_,
        a multivalid conformal prediction method that satisfies coverage guarantees conditioned on group membership
        and non-conformity threshold.

        Parameters
        ----------
        seed: int
            Random seed.
        """
        super().__init__(seed=seed)



from typing import Optional

import optax

from fortuna.typing import OptaxOptimizer


class CalibOptimizer:
    def __init__(
        self,
        method: Optional[OptaxOptimizer] = optax.adam(1e-2),
        n_epochs: int = 100,
    ):
        """
        An object to configure the optimization in the calibration process.

        Parameters
        ----------
        method: OptaxOptimizer
            An Optax optimizer.
        n_epochs: int
            Maximum number of epochs to run the calibration for.
        """
        self.method = method
        self.n_epochs = n_epochs



import logging
import pandas as pd
import boto3
import jax.numpy as jnp
import numpy as np
from typing import Union, Optional
import jax
from jax import vmap
from sklearn.metrics import log_loss, roc_auc_score
from tqdm import tqdm
from fortuna.conformal import BinaryClassificationMulticalibrator

Array = Union[np.ndarray, jax.Array]


class MaxCoverageFixedPrecisionBinaryClassificationCalibrator:
    def __init__(self):
        self._patches = dict()

    def calibrate(
        self,
        targets: Array,
        probs: Array,
        true_positive_precision_threshold: float,
        false_negative_precision_threshold: float,
        test_probs: Optional[Array] = None,
        n_taus: int = 100,
        margin: float = 0.
    ) -> Union[None, Array]:
        if false_negative_precision_threshold <= 0.5 or true_positive_precision_threshold <= 0.5:
            raise ValueError("Both `false_negative_precision_threshold` and"
                             " `true_positive_precision_threshold` must be greater than 0.5.")
        probs = jnp.copy(probs)
        targets = jnp.copy(targets)

        def _true_positive_objective_fn(tau: Array):
            calib_probs = jnp.clip((1 + (tau - 1) * (probs > 0.5)) * probs, a_max=1)
            b_pos_prec = calib_probs >= true_positive_precision_threshold
            prob_b_pos_prec = jnp.mean(b_pos_prec)
            pos_prec = jnp.mean(targets * b_pos_prec) / prob_b_pos_prec
            pos_cond = pos_prec >= true_positive_precision_threshold + margin
            return prob_b_pos_prec * pos_cond

        def _false_negative_objective_fn(tau: Array):
            calib_probs = (1 + (tau - 1) * (probs < 0.5)) * probs
            b_neg_prec = calib_probs <= 1 - false_negative_precision_threshold
            prob_b_neg_prec = jnp.mean(b_neg_prec)
            neg_prec = jnp.mean((1 - targets) * b_neg_prec) / prob_b_neg_prec
            neg_cond = neg_prec >= false_negative_precision_threshold + margin
            return prob_b_neg_prec * neg_cond

        taus_pos = jnp.linspace(1, 2 * true_positive_precision_threshold, n_taus)
        taus_neg = jnp.linspace(2 * (1 - false_negative_precision_threshold), 1, n_taus)[::-1]

        values_pos = vmap(_true_positive_objective_fn)(taus_pos)

        msg = "The {} could not be satisfied. Please consider improving the classifier or decreasing the threshold."

        if jnp.max(values_pos) == 0:
            logging.warning(msg.format("`true_positive_precision_threshold`"))
        values_neg = vmap(_false_negative_objective_fn)(taus_neg)
        if jnp.max(values_neg) == 0:
            logging.warning(msg.format("`false_negative_precision_threshold`"))

        self._patches["tau_pos"] = taus_pos[jnp.argmax(values_pos)]
        self._patches["tau_neg"] = taus_neg[jnp.argmax(values_neg)]

        if test_probs is not None:
            return self.apply_patches(test_probs)

    def apply_patches(
        self,
        probs: Array
    ) -> Array:
        if not len(self._patches):
            logging.warning("No patches available.")
            return probs

        probs = jnp.copy(probs)
        probs = probs.at[probs > 0.5].set(jnp.clip(self._patches["tau_pos"] * probs[probs > 0.5], a_max=1))
        probs = probs.at[probs < 0.5].set(self._patches["tau_neg"] * probs[probs < 0.5])
        return probs

    @staticmethod
    def true_positive_precision(probs: Array, targets: Array, threshold: float):
        b = probs >= threshold
        prob_b = jnp.mean(b)
        return jnp.mean(targets * b) / prob_b

    @staticmethod
    def false_negative_precision(probs: Array, targets: Array, threshold: float):
        b = probs <= 1 - threshold
        prob_b = jnp.mean(b)
        return jnp.mean((1 - targets) * b) / prob_b

    @staticmethod
    def true_positive_coverage(probs: Array, threshold: float):
        return jnp.mean(probs >= threshold)

    @staticmethod
    def false_negative_coverage(probs: Array, threshold: float):
        return jnp.mean(probs <= threshold)

    @property
    def patches(self):
        return self._patches


def get_precision(df, value):
    if len(df) == 0:
        return None
    correct_count = df.loc[df["decision"] == value]
    return len(correct_count)/len(df)


def get_tn_tp_performance(tp_threshold, tn_threshold, df, column_name):
    # decision is the column for human labels
    tp_df = df.loc[df[column_name]>=tp_threshold][["decision"]]
    tn_df = df.loc[df[column_name]<=tn_threshold][["decision"]]
    return get_precision(tp_df, 1), len(tp_df)/len(df), get_precision(tn_df, 0), len(tn_df)/len(df)


def get_result(class_name: str, probs: Array, pred_results: pd.DataFrame, column_name: str):
    calibrator_result = pd.concat([pred_results, pd.DataFrame(probs)], axis=1).rename(columns={0: column_name})

    tp_precision, tp_percent, tn_precision, tn_percent = get_tn_tp_performance(target_tp_threshold,
                                                                               1 - target_tn_threshold,
                                                                               calibrator_result,
                                                                               column_name)
    unique_decision = set(calibrator_result['decision'])
    if len(unique_decision) == 2:
        roc_score = roc_auc_score(calibrator_result['decision'], calibrator_result[column_name])
        logloss = log_loss(calibrator_result["decision"], calibrator_result[column_name])
    else:
        roc_score = float('nan')
        logloss = float('nan')

    total_count = len(calibrator_result)
    result = {}
    result["class"] = class_name
    result["roc_auc"] = roc_score
    result["logloss"] = logloss
    result["tp_precision"] = tp_precision
    result["tp_percent"] = tp_percent
    result["tn_precision"] = tn_precision
    result["tn_percent"] = tn_percent
    result["total_count"] = total_count
    result["total_coverage"] = (tp_percent if tp_precision is not None and round(tp_precision,
                                                                                 2) >= target_tp_threshold else 0) + \
                               (tn_percent if tn_precision is not None and round(tn_precision,
                                                                                 2) >= target_tn_threshold else 0)
    result["total_ar"] = result["total_coverage"] * total_count
    return pd.DataFrame.from_dict(result, orient='index').T


def fix_coverage(class_list, target_tp_threshold, target_tn_threshold, data_type, column_name):
    before_result_list = []
    after_result_list = []

    for class_name in tqdm(class_list):
        class_data_path = common_path + class_name
        try:
            dataset = pd.read_parquet(class_data_path)
        except:
            continue
        dataset["decision"] = dataset['MEMBER'].map({'true': 1, 'false': 0})
        dataset = dataset.drop(columns=["MEMBER"])

        dataset_cal = dataset[dataset['data_type'] == "cal"]
        cols_cal = ['item_id', 'decision', 'predicition_1']

        # the original val_results contains prediction results for both validation and calibration sets, we extract the outputs on calibration data by joining the results with calibration data.
        try:
            val_results = pd.read_parquet(
                f"s3://cpp-adc-data-sharing/model_outputs_sample_classes/{class_name}/val-sgs/result.parquet").rename(
                columns={"prediciton_val-sgs": "predicition_1"})
            calibration_results = pd.merge(dataset_cal.drop("decision", axis=1), val_results, on="item_id", how='inner')[
                cols_cal]
        except:
            continue

        dataset_type = dataset[dataset['data_type'] == data_type]
        cols = ['item_id', 'decision', 'predicition_1']

        # validation data
        if data_type == "val":
            results = val_results
        # test data
        else:
            results = pd.read_parquet(
                f"s3://cpp-adc-data-sharing/model_outputs_sample_classes/{class_name}/test-sgs/result.parquet").rename(
                columns={"prediciton_test-sgs": "predicition_1"})

        pred_results = pd.merge(dataset_type.drop("decision", axis=1), results, on="item_id", how='inner')[cols]

        calibration_groups = pd.DataFrame({
            'greater_than_0.9': calibration_results['predicition_1'] > 0.9,
            'less_than_0.1': calibration_results['predicition_1'] < 0.1
            # 'in_the_middle': (calibration_results['predicition_1'] >= 0.1) & (calibration_results['predicition_1'] < 0.9)
        })

        groups = pd.DataFrame({
            'greater_than_0.9': pred_results['predicition_1'] > 0.9,
            'less_than_0.1': pred_results['predicition_1'] < 0.1
            # 'in_the_middle': (pred_results['predicition_1'] >= 0.1) & (pred_results['predicition_1'] < 0.9)
        })

        mc = BinaryClassificationMulticalibrator()
        status = mc.calibrate(
            targets=calibration_results['decision'].values,
            probs=calibration_results['predicition_1'].values,
            n_buckets=150,
            groups=calibration_groups.values,
        )

        mc_calib_probs = mc.apply_patches(groups=calibration_groups.values, probs=calibration_results['predicition_1'].values)
        mc_test_probs = mc.apply_patches(groups=groups.values, probs=pred_results['predicition_1'].values)

        calibrator = MaxCoverageFixedPrecisionBinaryClassificationCalibrator()
        new_test_probs = calibrator.calibrate(
            targets=calibration_results['decision'].values,
            probs=mc_calib_probs,
            test_probs=mc_test_probs,
            true_positive_precision_threshold=target_tp_threshold,
            false_negative_precision_threshold=target_tn_threshold,
            margin=0
        )

        before_result_list.append(
            get_result(
                class_name=class_name,
                probs=pred_results['predicition_1'].values,
                pred_results=pred_results,
                column_name=column_name
            )
        )
        after_result_list.append(
            get_result(
                class_name=class_name,
                probs=new_test_probs,
                pred_results=pred_results,
                column_name=column_name
            )
        )
    before_result_df = pd.concat(before_result_list)
    after_result_df = pd.concat(after_result_list)

    return before_result_df, after_result_df


if __name__ == "__main__":
    sm_role = f"arn:aws:iam::848675773966:role/admin"
    boto_sess = boto3.Session(profile_name="default", region_name="us-east-1")

    class_name = "%SC%ProductSafety%Regulated%SG_Product_Assurance_Toys_Games_and_Sports_Activities_HR"
    common_path = 's3://cpp-adc-data-sharing/ar_benchmark_dataset/'
    class_data_path = common_path + class_name
    dataset = pd.read_parquet(class_data_path)

    target_tp_threshold = 0.97
    target_tn_threshold = 0.99
    data_type = "val"

    class_list = [
        "%SC%DIPC%AU%Art_Supplies_General_Use_AU_Wooden_Pencils_Art_Supplies_General_Use",
        "%SC%DIPC%CA%Apparel_Accessories_US_CA_ITK_Headwear_Adult",
        "%SC%DIPC%CA%BISS_General_Use_US_CA_Raw_Materials_General_Use",
        "%SC%DIPC%CA%Consumer_Electrical_Products_US_CA_Electrical_Vacuum_Sealers_ML",
        "%SC%DIPC%CA%DVDs_US_CA_ITK_Magic_DVDs_General_Use",
        "%SC%DIPC%CA%General_Use_US_CA_Garden_Tools_Wooden_ML",
        "%SC%DIPC%CA%General_Use_US_CA_Self_Locking_Security_Seals_Tags_General_Use_ML",
        "%SC%DIPC%CA%Home_General_Use_US_CA_Subcat_Vacuum_Accessories_and_Replacement_Parts",
        "%SC%DIPC%CA%Textiles_General_Use_US_CA_Textile_Table_Runners_ML",
        "%SC%DIPC%JP%MEDIA_General_Use_JP_GL_DVD",
        "%SC%DIPC%MX%Labeling_General_Rugs_Mats",
        "%SC%DIPC%SG%SG_Cosmetics_General_Use_SG_Mouthwash_Toothpaste_ML",
        "%SC%DIPC%UK%Apparel_General_Use_EU_ITK_Apparel_Adult",
        "%SC%DIPC%UK%Auto_General_Use_EU_Car_and_Car_Seat_Covers",
        "%SC%DIPC%UK%CE_General_Use_EU_Cables_without_Adapter",
        "%SC%DIPC%UK%Chemicals_Unsupported_EU_Self_Inking_Stamps",
        "%SC%DIPC%UK%Feeding_Bottle_Nipples_Infant_EU_ITK_Baby_Bottle_Nipples_Infant",
        "%SC%DIPC%UK%General_Use_EU_Masonry_Floats",
        "%SC%DIPC%UK%Jewelry_General_Use_Precious_EU_Jewelry_Precious_General_Use",
        "%SC%ProductSafety%EURsP%DE_Electronics_Hygrometer_ML",
        "%SC%ProductSafety%EURsP%FR_Electronics_Electrical_Housings_ML",
        "%SC%ProductSafety%EURsP%FR_Electronics_Pruning_Shrears_ML_TB",
        "%SC%ProductSafety%EURsP%IT_Electronics_Electric_Blanket_ML",
        "%SC%ProductSafety%Illegal %IT_Product_Safety_Angle_Grinder_Chainsaw_Disc",
        "%SC%ProductSafety%RecalledProducts%AE_Product_Safety_Self_Balancing_Scooters",
        "%SC%ProductSafety%RecalledProducts%EU_Recalls_ATD_ES_Mermaid_tail",
        "%SC%ProductSafety%RecalledProducts%EU_Recalls_ES_Christmas_Dancing_Cactus",
        "%SC%ProductSafety%RecalledProducts%EU_Recalls_RFR_DE_LED_Christmas_Hat",
        "%SC%ProductSafety%RecalledProducts%EU_Recalls_RFR_FR_Kinderslot_Baby_lock",
        "%SC%ProductSafety%RecalledProducts%EU_Recalls_UK_Christmas_Dancing_Cactus",
        "%SC%ProductSafety%Regulated%AU_Product_Assurance_Toy_Action_Figures_And_Dolls_AutoML",
        "%SC%ProductSafety%Regulated%CA_Product_Assurance_Childrens_Jewellery_AutoML",
        "%SC%ProductSafety%Regulated%CA_Product_Assurance_Free_Standing_Clothing_Storage_Units_V00_HR_New",
        "%SC%ProductSafety%Regulated%CA_Product_Assurance_Gates_And_Enclosures_HR",
        "%SC%ProductSafety%Regulated%FR_Product_Assurance_Toys_Balloons_and_Party_Toys_HRE",
        "%SC%ProductSafety%Regulated%FR_Toys_Dolls_Dollhouses_and_Accessories_AutoML",
        "%SC%ProductSafety%Regulated%FR_Toys_Kitchen_Food_AutoML_PA",
        "%SC%ProductSafety%Regulated%IN_Product_Assurance_Toys_Games_And_Puzzles_AutoML",
        "%SC%ProductSafety%Regulated%JP_Product_Assurance_Toys_Art_and_Craft_AutoML2",
        "%SC%ProductSafety%Regulated%JP_Product_Assurance_Toys_Lawn_and_Gardening_Sets_and_Tool_kits_Playground_Equipment_AutoML2",
        "%SC%ProductSafety%Regulated%SG_Product_Assurance_Toys_Games_and_Sports_Activities_HR",
        "%SC%ProductSafety%Regulated%UK_Product_Assurance_Toys_Ride_Ons_HR",
        "%SC%ProductSafety%Regulated%US_Product_Assurance_Carriages_And_Strollers_HR",
        "%SC%ProductSafety%Regulated%US_Product_Assurance_Trampolines_HR",
        "%SC%ProductSafety%RegulatoryEngagement %US_Multiple_BathroomShelves",
        "%SC%ProductSafety%UPSC%AU_Baby_Feeding_ML_Path",
        "%SC%ProductSafety%UPSC%AU_Sunglasses_ML_Path",
        "%SC%ProductSafety%UPSC%CA_Car_Seat_Heating_Pads_ML_No_Path",
        "%SC%ProductSafety%UPSC%ES_ESPN_Beaded_Teethers_ML_No_Path",
        "%SC%ProductSafety%UPSC%ES_ESPN_Padded_Crib_Bumpers_ML_No_Path",
        "%SC%ProductSafety%UPSC%MX_Lithiumn_Cylindrical_Battery_Cell_KW_No_Path",
        "%SC%ProductSafety%UPSC%PL_Padded_Crib_Bumpers_KW_No_Path",
        "%SC%ProductSafety%UPSC%SG_SGPN_Standalone_Power_Banks_ML_Path",
        "%SC%ProductSafety%UPSC%UK_UKPN_Costume_Sets_ML_Path_Website_Compliance",
        "%SC%ProductSafety%UPSC%US_USPN_Childrens_Jewelry_KW_Path",
        "%SC%ProductSafety%UPSC%US_USPN_Laptop_Chargers_ML_Path",
        "%SC%ProductSafety%UPSC%US_USPN_Squishy_Toys_ML_Path",
        "%SC%RestrictedProducts%AmazonPolicy%BE_Sex_Products_Realistic_Sex_Dolls_FRKW",
        "%SC%RestrictedProducts%AmazonPolicy%DE_Party_Liquor",
        "%SC%RestrictedProducts%AmazonPolicy%Drug_Paraphernalia_Marijuana_Grinders",
        "%SC%RestrictedProducts%AmazonPolicy%ES_Mystery_Boxes_ML",
        "%SC%RestrictedProducts%AmazonPolicy%FR_Mystery_Boxes_ML",
        "%SC%RestrictedProducts%AmazonPolicy%IN_Improper_Listing_Of_Food_Suppress",
        "%SC%RestrictedProducts%AmazonPolicy%IT_Mystery_Boxes_ML",
        "%SC%RestrictedProducts%AmazonPolicy%PLNT_Christmas_Trees_1",
        "%SC%RestrictedProducts%AmazonPolicy%SE_Currency_Legal_Tender_Yen",
        "%SC%RestrictedProducts%AmazonPolicy%TR_Sex_Products_Erotic_Dices_Cards",
        "%SC%RestrictedProducts%AmazonPolicy%UK_Animals_Fur_Products_Rabbit_II",
        "%SC%RestrictedProducts%AmazonPolicy%Weapons_Stun_Gun",
        "%SC%RestrictedProducts%GatedProducts%ES_Corrective_Contact_Lenses_Gating_ML",
        "%SC%RestrictedProducts%GatedProducts%Softlines_Fur_Baby_Outdoor_Sports_Retail_Removal",
        "%SC%RestrictedProducts%GatedProducts%TR_Medical_Devices_Launch_PTs_ID",
        "%SC%RestrictedProducts%GatedProducts%Textile_Bamboo_Antimicrobial_Retail_Removal_Clothing_ML",
        "%SC%RestrictedProducts%GatedProducts%Textile_Bamboo_Retail_Removal",
        "%SC%RestrictedProducts%Illegal%Auto_Motorcycle_Helmets_DOT",
        "%SC%RestrictedProducts%Illegal%Auto_Tuner_Claims",
        "%SC%RestrictedProducts%Illegal%CA_Medical_Device_Hearing_Aids",
        "%SC%RestrictedProducts%Illegal%Currency_US_Paper_Currency",
        "%SC%RestrictedProducts%Illegal%DE_Fireworks_Fire_Extinguisher_Ball",
        "%SC%RestrictedProducts%Illegal%EG_Weapons_Firearm_Accessories",
        "%SC%RestrictedProducts%Illegal%JP_Illegal_Obscene_Sex_Dolls_And_Toys",
        "%SC%RestrictedProducts%Illegal%Medical_Device_Electrosurgical",
        "%SC%RestrictedProducts%Illegal%PL_Magnets_Buckyballs",
        "%SC%RestrictedProducts%Illegal%SA_Electronics_Two_Way_Radio",
        "%SC%RestrictedProducts%Illegal%SA_Sexual_Wellness_Products_ALL",
        "%SC%RestrictedProducts%Illegal%UK_Cigarette_Novelty_Lighters",
        "%SC%RestrictedProducts%Illegal%UK_Weapons_Knives_Missing_CRT_Flag_SfT_HR_No_Policy",
        "%SC%RestrictedProducts%Illegal%Unapproved_Drug_Skin_Patches",
        "%SC%RestrictedProducts%IllegallyMkted%SE_Adult_Products_With_Childlike_Features",
        "%SC%RestrictedProducts%IllegallyMkted%UK_Tobacco_E-Cigarettes_Without_Flag_Suppress_SfT_HR_No_Policy_ML",
        "%SC%RestrictedProducts%IllegallyMkted%UK_Weapons_Commando_Knives_SfT_No_Policy_ML",
        "%SC%RestrictedProducts%Regulated%BR_Medical_Devices_Tattoo_Equipment",
        "%SC%RestrictedProducts%Regulated%CARB205_Exhaust_Header_Identification_ML",
        "%SC%RestrictedProducts%Regulated%DE_Electronics_E-Mobility_Identify_ML",
        "%SC%RestrictedProducts%Regulated%IN_Class_A_NSNM_Medical_Devices_2",
        "%SC%RestrictedProducts%Regulated%IN_Class_A_NSNM_Medical_Devices_9",
        "%SC%RestrictedProducts%Regulated%TR_Electronics_E-Mobility_Identify_ML"
]

    all_names = []
    for class_name in tqdm.tqdm(class_list):
        try:
            dataset = pd.read_parquet(class_data_path)
            val_results = pd.read_parquet(
                f"s3://cpp-adc-data-sharing/model_outputs_sample_classes/{class_name}/val-sgs/result.parquet").rename(
                columns={"prediciton_val-sgs": "predicition_1"})
            all_names.append(class_name)
        except:
            continue

    print(all_names)



import abc
import logging
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
    Type,
)

from flax.core import FrozenDict
from jax._src.prng import PRNGKeyArray

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.posterior_mixin import WithPosteriorCheckpointingMixin
from fortuna.prob_model.posterior.posterior_state_repository import (
    PosteriorStateRepository,
)
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import (
    Path,
    Status,
)
from fortuna.utils.freeze import get_trainable_paths
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)
from fortuna.utils.random import WithRNG


class PosteriorApproximator(abc.ABC):
    """
    A posterior approximator abstract class.
    """

    @abc.abstractmethod
    def __str__(self):
        pass

    @property
    def posterior_method_kwargs(self) -> Dict[str, Any]:
        return {}


class Posterior(WithRNG, WithPosteriorCheckpointingMixin):
    state = None

    def __init__(self, joint: Joint, posterior_approximator: PosteriorApproximator):
        r"""
        Posterior distribution class. This refers to :math:`p(w|\mathcal{D}, \phi)`, where :math:`w` are the random
        model parameters, :math:`\mathcal{D}` is a training data set and :math:`\phi` are calibration parameters.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: PosteriorApproximator
            A posterior approximator.
        """
        super().__init__()
        self.joint = joint
        self.posterior_approximator = posterior_approximator

    def _restore_state_from_somewhere(
        self,
        fit_config: FitConfig,
        allowed_states: Optional[Tuple[Type[PosteriorState], ...]] = None,
    ) -> PosteriorState:
        if fit_config.checkpointer.restore_checkpoint_path is not None:
            state = self.restore_checkpoint(
                restore_checkpoint_path=fit_config.checkpointer.restore_checkpoint_path,
                optimizer=fit_config.optimizer.method,
            )
        elif fit_config.checkpointer.start_from_current_state is not None:
            state = self.state.get(optimizer=fit_config.optimizer.method)

        if allowed_states is not None and not isinstance(state, allowed_states):
            raise ValueError(
                f"The type of the restored checkpoint must be within {allowed_states}. "
                f"However, {fit_config.checkpointer.restore_checkpoint_path} pointed to a state "
                f"with type {type(state)}."
            )

        return state

    def _init_joint_state(self, data_loader: DataLoader) -> JointState:
        return self.joint.init(input_shape=data_loader.input_shape)

    @staticmethod
    def _freeze_optimizer_in_state(
        state: PosteriorState, fit_config: FitConfig
    ) -> PosteriorState:
        if fit_config.optimizer.freeze_fun is not None:
            trainable_paths = get_trainable_paths(
                state.params, fit_config.optimizer.freeze_fun
            )
            state = state.replace(
                opt_state=fit_config.optimizer.method.init(
                    FrozenDict(
                        nested_set(
                            d={},
                            key_paths=trainable_paths,
                            objs=tuple(
                                [
                                    nested_get(state.params.unfreeze(), path)
                                    for path in trainable_paths
                                ]
                            ),
                            allow_nonexistent=True,
                        )
                    )
                )
            )
        return state

    @staticmethod
    def _check_state(state: PosteriorState) -> None:
        pass

    @abc.abstractmethod
    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        **kwargs,
    ) -> Status:
        """
        Fit the posterior distribution. A posterior state will be internally stored.

        Parameters
        ----------
        train_data_loader: DataLoader
            Training data loader.
        val_data_loader: Optional[DataLoader]
            Validation data loader.
        fit_config: FitConfig
            A configuration object.

        Returns
        -------
        Status
            A status including metrics describing the fitting process.
        """
        pass

    @abc.abstractmethod
    def sample(self, rng: Optional[PRNGKeyArray] = None, *args, **kwargs) -> JointState:
        """
        Sample from the posterior distribution.

        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        pass

    def load_state(self, checkpoint_path: Path) -> None:
        """
        Load the state of the posterior distribution from a checkpoint path. The checkpoint must be
        compatible with the current probabilistic model.

        Parameters
        ----------
        checkpoint_path: Path
            Path to checkpoint file or directory to restore.
        """
        try:
            self.restore_checkpoint(checkpoint_path)
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_path={checkpoint_path}`."
            )
        self.state = PosteriorStateRepository(checkpoint_dir=checkpoint_path)

    def save_state(
        self, checkpoint_path: Path, keep_top_n_checkpoints: int = 1
    ) -> None:
        """
        Save the state of the posterior distribution to a checkpoint directory.

        Parameters
        ----------
        checkpoint_path: Path
            Path to checkpoint file or directory to restore.
        keep_top_n_checkpoints: int
            Number of past checkpoint files to keep.
        """
        if self.state is None:
            raise ValueError(
                """No state available. You must first either fit the posterior distribution, or load a
            saved checkpoint."""
            )
        return self.state.put(
            self.state.get(),
            checkpoint_path=checkpoint_path,
            keep=keep_top_n_checkpoints,
        )

    def _check_fit_config(self, fit_config: FitConfig):
        if (
            fit_config.checkpointer.dump_state is True
            and not fit_config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )

    @staticmethod
    def _is_state_available_somewhere(fit_config: FitConfig) -> bool:
        return (
            fit_config.checkpointer.restore_checkpoint_path is not None
            or fit_config.checkpointer.start_from_current_state
        )

    def _warn_frozen_params_start_from_random(
        self, fit_config: FitConfig, map_fit_config: Optional[FitConfig]
    ) -> None:
        if (
            not self._is_state_available_somewhere(fit_config)
            and map_fit_config is None
            and fit_config.optimizer.freeze_fun is not None
        ):
            logging.warning(
                "Parameters frozen via `fit_config.optimizer.freeze_fun` will not be updated. To start "
                "from sensible frozen parameters, you should configure "
                "`fit_config.checkpointer.restore_checkpoint_path`, or "
                "`fit_config.checkpointer.start_from_current_state`, or `map_fit_config`. "
                "Otherwise, "
                "a randomly initialized configuration of frozen parameters will be returned."
            )

    def _checks_on_fit_start(
        self, fit_config: FitConfig, map_fit_config: Optional[FitConfig]
    ) -> None:
        self._check_fit_config(fit_config)
        self._warn_frozen_params_start_from_random(fit_config, map_fit_config)

    def _should_run_preliminary_map(
        self, fit_config: FitConfig, map_fit_config: Optional[FitConfig]
    ) -> bool:
        return (
            not self._is_state_available_somewhere(fit_config)
            and map_fit_config is not None
        )



import jax.numpy as jnp
from jax.tree_util import tree_map
import numpy as np

from fortuna.data.loader import DataLoader
from fortuna.typing import (
    InputData,
    Shape,
)


def check_data_loader_is_not_random(data_loader: DataLoader, max_iter: int = 3) -> None:
    flag = False
    for i, ((x1, y1), (x2, y2)) in enumerate(zip(data_loader, data_loader)):
        if i > max_iter:
            break
        if isinstance(x1, dict):
            if not all(
                [np.alltrue(x1[key] == x2[key]) for key in x1.keys()]
            ) or not np.alltrue(y1 == y2):
                flag = True
                break
        else:
            if not np.alltrue(x1 == x2) or not np.alltrue(y1 == y2):
                flag = True
                break

    if flag:
        raise ValueError(
            """The data loader randomizes at every iteration. To perform this method, please provide a data loader that
            generates the same sequence of data when called multiple times."""
        )


def get_input_shape(inputs: InputData) -> Shape:
    return tree_map(lambda x: x.shape[1:], inputs)


def get_inputs_from_shape(input_shape: Shape) -> InputData:
    if isinstance(input_shape, tuple):
        inputs = jnp.zeros((1,) + input_shape)
    elif isinstance(input_shape, dict):
        inputs = {k: jnp.zeros((1,) + v) for k, v in input_shape.items()}
    else:
        raise ValueError("Data batches shape have to be of type fortuna.typing.Shape")
    return inputs



from fortuna.prob_model.calib_config.base import CalibConfig
from fortuna.prob_model.calib_config.checkpointer import CalibCheckpointer
from fortuna.prob_model.calib_config.monitor import CalibMonitor
from fortuna.prob_model.calib_config.optimizer import CalibOptimizer
from fortuna.prob_model.calib_config.processor import CalibProcessor
from fortuna.prob_model.classification import ProbClassifier
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.fit_config.checkpointer import FitCheckpointer
from fortuna.prob_model.fit_config.monitor import FitMonitor
from fortuna.prob_model.fit_config.optimizer import FitOptimizer
from fortuna.prob_model.fit_config.processor import FitProcessor
from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_approximator import (
    DeepEnsemblePosteriorApproximator,
)
from fortuna.prob_model.posterior.laplace.laplace_approximator import (
    LaplacePosteriorApproximator,
)
from fortuna.prob_model.posterior.map.map_posterior import MAPPosteriorApproximator
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_approximator import (
    ADVIPosteriorApproximator,
)
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld.cyclical_sgld_approximator import (
    CyclicalSGLDPosteriorApproximator,
)
from fortuna.prob_model.posterior.sgmcmc.sghmc.sghmc_approximator import (
    SGHMCPosteriorApproximator,
)
from fortuna.prob_model.posterior.sngp.sngp_approximator import (
    SNGPPosteriorApproximator,
)
from fortuna.prob_model.posterior.swag.swag_approximator import (
    SWAGPosteriorApproximator,
)
from fortuna.prob_model.regression import ProbRegressor



import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array


class RegressionTemperatureScaler(nn.Module):
    r"""
    Regression temperature scaling. It multiplies the variance with a scalar temperature parameters. Let :math:`v` be
    the variance outputs and :math:`\phi` be a scalar parameter. Then the scaling can be seen as
    :math:`g(\phi, o) = \exp(\phi) v`.
    """

    @nn.compact
    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        log_temp = self.param("log_temp", nn.initializers.zeros, (1,))
        mean, log_var = jnp.split(x, 2, axis=-1)
        log_var += log_temp
        return jnp.concatenate((mean, log_var), axis=-1)



import logging
from typing import Type

import jax


def select_trainer_given_devices(
    devices: int,
    base_trainer_cls: Type,
    jitted_trainer_cls: Type,
    multi_device_trainer_cls: Type,
    disable_jit: bool,
) -> Type:
    if devices not in [0, -1]:
        raise NotImplementedError(
            "Currently, only two options are supported: use all available (`devices=-1`) or use only CPU (`devices=0`)."
        )
    elif devices == -1 and disable_jit:
        logging.warning("Jit must be enabled when not training on a single CPU device.")

    if devices == -1:
        logging.info("Training on all available devices.")
        trainer_cls = (
            multi_device_trainer_cls
            if len([d for d in jax.devices() if d.platform == "gpu"]) > 0
            else jitted_trainer_cls
        )

    elif devices == 0 and disable_jit:
        logging.info("Training on CPU without jit.")
        trainer_cls = base_trainer_cls
    else:
        logging.info("Training on CPU.")
        trainer_cls = jitted_trainer_cls
    return trainer_cls



from typing import (
    Dict,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
import flax.linen as nn
from flax.training.checkpoints import PyTree
import jax
from jax import random
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.model.model_manager.base import ModelManager
from fortuna.typing import (
    Array,
    Mutable,
    Params,
)
from fortuna.utils.data import get_inputs_from_shape


class RegressionModelManager(ModelManager):
    def __init__(
        self,
        model: nn.Module,
        likelihood_log_variance_model: nn.Module,
        model_editor: Optional[nn.Module] = None,
    ):
        r"""
        Regression model manager class. It orchestrates the forward pass of the model in the probabilistic model.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. It characterizes the mean model
            of the likelihood function. The outputs must belong to the same space as the target variables.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`\mu(w, x)`.
        likelihood_log_variance_model: nn.Module
            A model characterizing the log-variance of a Gaussian likelihood function. The outputs must belong to the
            same space as the target variables. Let :math:`x` be input variables and :math:`w` the random model
            parameters. Then the model is described by a function :math:`\log\sigma^2(w, x)`.
        """
        super(RegressionModelManager, self).__init__(model, model_editor)
        self.likelihood_log_variance_model = likelihood_log_variance_model

    def apply(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        train: bool = False,
        rng: Optional[PRNGKeyArray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        # setup dropout key
        if rng is not None:
            rng, model_dropout_key, lik_log_var_dropout_key = random.split(rng, 3)
            model_rngs = {"dropout": model_dropout_key}
            lik_log_var_rngs = {"dropout": lik_log_var_dropout_key}
        else:
            model_rngs = None
            lik_log_var_rngs = None

        if mutable is not None:
            mutable["model"] = mutable.get("model")
            mutable["lik_log_var"] = mutable.get("lik_log_var")

        model_has_aux = train and mutable is not None and mutable["model"] is not None
        lik_log_var_has_aux = (
            train and mutable is not None and mutable["lik_log_var"] is not None
        )

        def apply_fn(p, x, m_mutable, llv_mutable):
            model_outputs = self.model.apply(
                p["model"],
                x,
                train=train,
                mutable=m_mutable,
                rngs=model_rngs,
            )
            lik_log_var_outputs = self.likelihood_log_variance_model.apply(
                p["lik_log_var"],
                x,
                train=train,
                mutable=llv_mutable,
                rngs=lik_log_var_rngs,
            )

            if isinstance(model_outputs, tuple) and not model_has_aux:
                model_outputs = model_outputs[0]
            if isinstance(lik_log_var_outputs, tuple) and not lik_log_var_has_aux:
                lik_log_var_outputs = lik_log_var_outputs[0]

            if model_has_aux:
                model_outputs, m_mutable = model_outputs
            if lik_log_var_has_aux:
                lik_log_var_outputs, llv_mutable = lik_log_var_outputs

            self._check_outputs(model_outputs, lik_log_var_outputs)

            aux = dict()
            if m_mutable or llv_mutable:
                aux["mutable"] = dict()
                if m_mutable:
                    aux["mutable"]["model"] = m_mutable
                if llv_mutable:
                    aux["mutable"]["lik_log_var"] = m_mutable

            return jnp.concatenate(
                (model_outputs, lik_log_var_outputs), axis=-1
            ), FrozenDict(aux)

        if self.model_editor is not None:
            outputs, aux = self.model_editor.apply(
                params["model_editor"],
                apply_fn=lambda p, x: apply_fn(
                    p,
                    x,
                    m_mutable=mutable["model"] if mutable is not None else False,
                    llv_mutable=mutable["lik_log_var"]
                    if mutable is not None
                    else False,
                ),
                model_params=params,
                x=inputs,
                has_aux=True,
            )
        else:
            outputs, aux = apply_fn(
                params,
                inputs,
                m_mutable=mutable["model"] if mutable is not None else False,
                llv_mutable=mutable["lik_log_var"] if mutable is not None else False,
            )

        if len(aux) > 0:
            return outputs, aux
        return outputs

    def init(
        self, input_shape: Tuple, rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, FrozenDict]:
        if rng is None:
            rng = self.rng.get()
        (
            rng,
            model_params_key,
            model_dropout_key,
            lik_log_var_params_key,
            lik_log_var_dropout_key,
        ) = random.split(rng, 5)
        model_rngs = {"params": model_params_key, "dropout": model_dropout_key}
        lik_log_var_rngs = {
            "params": lik_log_var_params_key,
            "dropout": lik_log_var_params_key,
        }
        params = dict(
            model=self.model.init(model_rngs, jnp.zeros((1,) + input_shape), **kwargs),
            lik_log_var=self.likelihood_log_variance_model.init(
                lik_log_var_rngs, jnp.zeros((1,) + input_shape), **kwargs
            ),
        )

        def apply_fn(p, x):
            model_outputs = self.model.apply(
                p["model"],
                x,
                rngs=model_rngs,
            )
            lik_log_var_outputs = self.likelihood_log_variance_model.apply(
                p["lik_log_var"],
                x,
                rngs=lik_log_var_rngs,
            )

            self._check_outputs(model_outputs, lik_log_var_outputs)

            return jnp.concatenate((model_outputs, lik_log_var_outputs), axis=-1)

        if self.model_editor is not None:
            if rng is None:
                rng = self.rng
            rng, params_key, dropout_key = random.split(rng, 3)
            rngs = {"params": params_key, "dropout": dropout_key}
            params.update(
                dict(
                    model_editor=self.model_editor.init(
                        rngs,
                        apply_fn=apply_fn,
                        model_params=params,
                        x=get_inputs_from_shape(input_shape),
                        has_aux=False,
                    )
                )
            )
        return params

    def _check_outputs(
        self, model_outputs: jnp.ndarray, lik_log_var_outputs: jnp.ndarray
    ) -> None:
        if model_outputs.shape[-1] != lik_log_var_outputs.shape[-1]:
            raise ValueError(
                f"""The output dimensions of `model` and `likelihood_log_variance_model must be the same.
            However, {model_outputs.shape[-1]} and {lik_log_var_outputs.shape[-1]} were found, respectively."""
            )



import logging
import pathlib
from typing import Optional

from flax.core import FrozenDict

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.map.map_trainer import (
    JittedMAPTrainer,
    MAPTrainer,
    MultiDeviceMAPTrainer,
)
from fortuna.prob_model.posterior.run_preliminary_map import run_preliminary_map
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld import CYCLICAL_SGLD_NAME
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld.cyclical_sgld_approximator import (
    CyclicalSGLDPosteriorApproximator,
)
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld.cyclical_sgld_callback import (
    CyclicalSGLDSamplingCallback,
)
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld.cyclical_sgld_integrator import (
    cyclical_sgld_integrator,
)
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld.cyclical_sgld_state import (
    CyclicalSGLDState,
)
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_posterior import SGMCMCPosterior
from fortuna.prob_model.posterior.sgmcmc.sgmcmc_posterior_state_repository import (
    SGMCMCPosteriorStateRepository,
)
from fortuna.typing import Status
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.freeze import get_trainable_paths
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)

logger = logging.getLogger(__name__)


class CyclicalSGLDPosterior(SGMCMCPosterior):
    def __init__(
        self,
        joint: Joint,
        posterior_approximator: CyclicalSGLDPosteriorApproximator,
    ):
        """
        Cyclical Stochastic Gradient Langevin Dynamics (SGLD) approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A Joint distribution object.
        posterior_approximator: CyclicalSGLDPosteriorApproximator
            A cyclical SGLD posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return CYCLICAL_SGLD_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config: Optional[FitConfig] = None,
        **kwargs,
    ) -> Status:
        super()._checks_on_fit_start(fit_config, map_fit_config)

        status = {}

        map_state = None
        if map_fit_config is not None and fit_config.optimizer.freeze_fun is None:
            logging.warning(
                "It appears that you are trying to configure `map_fit_config`. "
                "However, a preliminary run with MAP is supported only if "
                "`fit_config.optimizer.freeze_fun` is given. "
                "Since the latter was not given, `map_fit_config` will be ignored."
            )
        elif not super()._is_state_available_somewhere(
            fit_config
        ) and super()._should_run_preliminary_map(fit_config, map_fit_config):
            map_state, status["map"] = run_preliminary_map(
                joint=self.joint,
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                map_fit_config=map_fit_config,
                rng=self.rng,
                **kwargs,
            )

        if fit_config.optimizer.method is not None:
            logging.info(f"`FitOptimizer` method in CyclicalSGLD is ignored.")

        fit_config.optimizer.method = cyclical_sgld_integrator(
            rng_key=self.rng.get(),
            init_step_size=self.posterior_approximator.init_step_size,
            cycle_length=self.posterior_approximator.cycle_length,
            exploration_ratio=self.posterior_approximator.exploration_ratio,
            preconditioner=self.posterior_approximator.preconditioner,
        )

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            base_trainer_cls=MAPTrainer,
            jitted_trainer_cls=JittedMAPTrainer,
            multi_device_trainer_cls=MultiDeviceMAPTrainer,
            disable_jit=fit_config.processor.disable_jit,
        )

        save_checkpoint_dir = (
            pathlib.Path(fit_config.checkpointer.save_checkpoint_dir) / "c"
            if fit_config.checkpointer.save_checkpoint_dir
            else None
        )
        trainer = trainer_cls(
            predict_fn=self.joint.likelihood.prob_output_layer.predict,
            save_checkpoint_dir=save_checkpoint_dir,
            save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
            eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
            early_stopping_monitor=fit_config.monitor.early_stopping_monitor,
            early_stopping_min_delta=fit_config.monitor.early_stopping_min_delta,
            early_stopping_patience=fit_config.monitor.early_stopping_patience,
            freeze_fun=fit_config.optimizer.freeze_fun,
        )

        if super()._is_state_available_somewhere(fit_config):
            state = self._restore_state_from_somewhere(fit_config=fit_config)
        else:
            state = self._init_map_state(map_state, train_data_loader, fit_config)

        if fit_config.optimizer.freeze_fun is not None:
            which_params = get_trainable_paths(
                params=state.params, freeze_fun=fit_config.optimizer.freeze_fun
            )
        else:
            which_params = None

        state = CyclicalSGLDState.convert_from_map_state(
            map_state=state,
            optimizer=fit_config.optimizer.method,
            which_params=which_params,
        )

        state = super()._freeze_optimizer_in_state(state, fit_config)

        self.state = SGMCMCPosteriorStateRepository(
            size=self.posterior_approximator.n_samples,
            checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
            which_params=which_params,
            all_params=state.params if which_params else None,
        )

        cyclical_sampling_callback = CyclicalSGLDSamplingCallback(
            n_epochs=fit_config.optimizer.n_epochs,
            n_training_steps=len(train_data_loader),
            n_samples=self.posterior_approximator.n_samples,
            n_thinning=self.posterior_approximator.n_thinning,
            cycle_length=self.posterior_approximator.cycle_length,
            exploration_ratio=self.posterior_approximator.exploration_ratio,
            trainer=trainer,
            state_repository=self.state,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
        )

        logging.info(f"Run CyclicalSGLD.")
        state, status = trainer.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=self.joint._batched_log_joint_prob,
            training_dataloader=train_data_loader,
            training_dataset_size=train_data_loader.size,
            n_epochs=fit_config.optimizer.n_epochs,
            metrics=fit_config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=val_data_loader.size
            if val_data_loader is not None
            else None,
            verbose=fit_config.monitor.verbose,
            callbacks=[cyclical_sampling_callback],
        )
        logging.info("Fit completed.")

        return status

    def _init_map_state(
        self,
        state: Optional[MAPState],
        data_loader: DataLoader,
        fit_config: FitConfig,
    ) -> MAPState:
        if state is None or fit_config.optimizer.freeze_fun is None:
            state = super()._init_joint_state(data_loader)

            return MAPState.init(
                params=state.params,
                mutable=state.mutable,
                optimizer=fit_config.optimizer.method,
                calib_params=state.calib_params,
                calib_mutable=state.calib_mutable,
            )
        else:
            random_state = super()._init_joint_state(data_loader)
            trainable_paths = get_trainable_paths(
                state.params, fit_config.optimizer.freeze_fun
            )
            state = state.replace(
                params=FrozenDict(
                    nested_set(
                        d=state.params.unfreeze(),
                        key_paths=trainable_paths,
                        objs=tuple(
                            [
                                nested_get(d=random_state.params, keys=path)
                                for path in trainable_paths
                            ]
                        ),
                    )
                )
            )

        return state



from typing import (
    List,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.likelihood.base import Likelihood
from fortuna.model.model_manager.state import ModelManagerState
from fortuna.output_calibrator.output_calib_manager.state import OutputCalibManagerState
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.prior.base import Prior
from fortuna.typing import (
    Batch,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
    Shape,
)
from fortuna.utils.data import get_inputs_from_shape
from fortuna.utils.random import WithRNG


class Joint(WithRNG):
    def __init__(self, prior: Prior, likelihood: Likelihood):
        r"""
        Joint distribution class. This is the joint distribution of target variables and random model parameters given
        input variables and calibration parameters. It is given by

        .. math::
            p(y, w|x, \phi),

        where:
         - :math:`x` is an observed input variable;
         - :math:`y` is an observed target variable;
         - :math:`w` denotes the random model parameters;
         - :math:`\phi` denotes the calibration parameters.

        Parameters
        ----------
        prior : Prior
            A prior distribution.
        likelihood : Likelihood
            A likelihood function.
        """
        self.prior = prior
        self.likelihood = likelihood

    def _batched_log_joint_prob(
        self,
        params: Params,
        batch: Batch,
        n_data: int,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        return_aux: Optional[List[str]] = None,
        train: Optional[bool] = False,
        outputs: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[float, Tuple[float, dict]]:
        """
        Evaluate the joint batched log-probability density function (a.k.a. log-pdf).

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        batch : Batch
            A batch of data points.
        n_data : int
            The total number of data points over which the likelihood is joint. This is used to rescale the batched
            log-likelihood function to better approximate the full likelihood.
        mutable : Optional[Mutable]
            The mutable objects used to evaluate the models.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable]
            The calibration mutable objects used to evaluate the calibrators.
        return_aux : Optional[List[str]]
            The auxiliary objects to return. We support 'outputs' and 'calib_mutable'. If this argument is not given,
            no auxiliary object is returned.
        train : Optional[bool]
            Whether the method is called during training.
        outputs : Optional[jnp.ndarray]
            Pre-computed batch of outputs.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        Union[float, Tuple[float, dict]]
            The evaluation of the joint batched log-pdf. If `return_aux` is given, the corresponding auxiliary objects
            are also returned.
        """
        if return_aux is None:
            return_aux = []
        log_prior = self.prior.log_joint_prob(params)
        outs = self.likelihood._batched_log_joint_prob(
            params,
            batch,
            n_data,
            mutable=mutable,
            calib_params=calib_params,
            calib_mutable=calib_mutable,
            return_aux=return_aux,
            train=train,
            outputs=outputs,
            rng=rng,
            **kwargs,
        )
        if len(return_aux) > 0:
            batched_log_lik, aux = outs
            return batched_log_lik + log_prior, aux
        batched_log_lik = outs
        return batched_log_lik + log_prior

    def _batched_negative_log_joint_prob(
        self,
        params: Params,
        batch: Batch,
        n_data: int,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        return_aux: Optional[List[str]] = None,
        train: Optional[bool] = False,
        outputs: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[float, Tuple[float, dict]]:
        outs = self._batched_log_joint_prob(
            params,
            batch,
            n_data,
            mutable,
            calib_params,
            calib_mutable,
            return_aux,
            train,
            outputs,
            rng,
            **kwargs,
        )
        if len(return_aux) > 0:
            loss, aux = outs
            loss *= -1
            return loss, aux
        return -outs

    def init(self, input_shape: Shape, **kwargs) -> JointState:
        """
        Initialize the state of the joint distribution.

        Parameters
        ----------
        input_shape : Shape
            The shape of the input variable.

        Returns
        -------
        A state of the joint distribution.
        """
        oms = ModelManagerState.init_from_dict(
            self.likelihood.model_manager.init(
                input_shape, rng=self.rng.get(), **kwargs
            )
        )
        inputs = get_inputs_from_shape(input_shape)
        outputs = self.likelihood.model_manager.apply(
            params=oms.params, inputs=inputs, mutable=oms.mutable
        )
        output_dim = (
            outputs[0].shape[-1]
            if isinstance(outputs, (list, tuple))
            else outputs.shape[-1]
        )
        ocms = OutputCalibManagerState.init_from_dict(
            FrozenDict(
                output_calibrator=self.likelihood.output_calib_manager.init(
                    output_dim=output_dim
                )
            )
        )
        return JointState.init_from_states(oms, ocms)



import abc
from typing import Optional

from jax._src.prng import PRNGKeyArray

from fortuna.typing import Params
from fortuna.utils.random import WithRNG


class Prior(WithRNG, abc.ABC):
    """
    Abstract prior distribution class.
    """

    @abc.abstractmethod
    def log_joint_prob(self, params: Params) -> float:
        """
        Evaluate the prior log-probability density function (a.k.a. log-pdf).

        Parameters
        ----------
        params : PyTree
            The parameters where to evaluate the log-pdf.

        Returns
        -------
        float
            Evaluation of the prior log-pdf.
        """
        pass

    @abc.abstractmethod
    def sample(self, params_like: Params, rng: Optional[PRNGKeyArray] = None) -> Params:
        """
        Sample parameters from the prior distribution.

        Parameters
        ----------
        params_like : PyTree
            An PyTree object with the same structure as the parameters to sample.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        PyTree
            A sample from the prior distribution.
        """
        pass



from typing import Optional


class FitHyperparameters:
    def __init__(
        self,
        max_grad_norm: Optional[float] = None,
        gradient_accumulation_steps: Optional[int] = None,
    ):
        """
        An object to configure additional arguments that may be needed during the posterior fitting.

        Parameters
        ----------
        max_grad_norm: Optional[Path]
            Maximum gradient norm. If `max_grad_norm > 0`, gradient clipping is performed.
        gradient_accumulation_steps: Optional[Path]
            Number of forward passes to perform before doing a backward pass.
        """
        self.max_grad_norm = max_grad_norm
        self.gradient_accumulation_steps = gradient_accumulation_steps



import abc
import logging
from typing import (
    Callable,
    Optional,
)

from flax.core import FrozenDict
import jax.numpy as jnp

from fortuna.calib_model.calib_mixin import WithCalibCheckpointingMixin
from fortuna.calib_model.calib_model_calibrator import (
    CalibModelCalibrator,
    JittedCalibModelCalibrator,
    MultiDeviceCalibModelCalibrator,
)
from fortuna.calib_model.calib_state_repository import CalibStateRepository
from fortuna.calib_model.config.base import Config
from fortuna.calib_model.loss import Loss
from fortuna.calib_model.state import CalibState
from fortuna.data.loader import DataLoader
from fortuna.model.model_manager.state import ModelManagerState
from fortuna.typing import (
    Outputs,
    Path,
    Predictions,
    Shape,
    Status,
    Targets,
    Uncertainties,
)
from fortuna.utils.data import (
    get_input_shape,
    get_inputs_from_shape,
)
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.freeze import get_trainable_paths
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)
from fortuna.utils.random import RandomNumberGenerator


class CalibModel(WithCalibCheckpointingMixin, abc.ABC):
    def __init__(self, seed: int = 0):
        """
        A calibration model.

        Parameters
        ----------
        seed: int = 0
            Random seed.
        """
        super().__init__()
        self.rng = RandomNumberGenerator(seed=seed)
        self.__set_rng()

    def __set_rng(self):
        self.model_manager.rng = self.rng
        self.prob_output_layer.rng = self.rng
        self.likelihood.rng = self.rng
        self.predictive.rng = self.rng

    def _calibrate(
        self,
        calib_data_loader: DataLoader,
        uncertainty_fn: Callable[[Predictions, Uncertainties, Targets], jnp.ndarray],
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray],
        val_data_loader: Optional[DataLoader] = None,
        config: Config = Config(),
    ) -> Status:
        if (
            config.checkpointer.dump_state is True
            and not config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )

        trainer_cls = select_trainer_given_devices(
            devices=config.processor.devices,
            base_trainer_cls=CalibModelCalibrator,
            jitted_trainer_cls=JittedCalibModelCalibrator,
            multi_device_trainer_cls=MultiDeviceCalibModelCalibrator,
            disable_jit=config.processor.disable_jit,
        )

        trainer = trainer_cls(
            predict_fn=self.prob_output_layer.predict,
            uncertainty_fn=uncertainty_fn,
            save_checkpoint_dir=config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=config.monitor.disable_calibration_metrics_computation,
            eval_every_n_epochs=config.monitor.eval_every_n_epochs,
            early_stopping_monitor=config.monitor.early_stopping_monitor,
            early_stopping_min_delta=config.monitor.early_stopping_min_delta,
            early_stopping_patience=config.monitor.early_stopping_patience,
            freeze_fun=config.optimizer.freeze_fun,
        )

        state = self._init_state(calib_data_loader, config)

        if config.optimizer.freeze_fun is not None:
            trainable_paths = get_trainable_paths(
                state.params, config.optimizer.freeze_fun
            )
            state = state.replace(
                opt_state=config.optimizer.method.init(
                    FrozenDict(
                        nested_set(
                            d={},
                            key_paths=trainable_paths,
                            objs=tuple(
                                [
                                    nested_get(state.params.unfreeze(), path)
                                    for path in trainable_paths
                                ]
                            ),
                            allow_nonexistent=True,
                        )
                    )
                )
            )

        loss = Loss(self.likelihood, loss_fn=loss_fn)
        loss.rng = self.rng

        n_calib_data = calib_data_loader.size
        n_val_data = val_data_loader.size if val_data_loader is not None else None

        if config.monitor.verbose:
            logging.info("Start calibration.")

        state, status = trainer.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=loss,
            training_dataloader=calib_data_loader,
            training_dataset_size=n_calib_data,
            n_epochs=config.optimizer.n_epochs,
            metrics=config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=n_val_data,
            verbose=config.monitor.verbose,
            callbacks=config.callbacks,
        )
        self.predictive.state = CalibStateRepository(
            config.checkpointer.save_checkpoint_dir
            if config.checkpointer.dump_state is True
            else None
        )
        self.predictive.state.put(
            state, keep=config.checkpointer.keep_top_n_checkpoints
        )
        if config.monitor.verbose:
            logging.info("Calibration completed.")
        return status

    def load_state(self, checkpoint_path: Path) -> None:
        """
        Load the state of the posterior distribution from a checkpoint path. The checkpoint must be compatible with the
        probabilistic model.

        Parameters
        ----------
        checkpoint_path : Path
            Path to a checkpoint file or directory to restore.
        """
        try:
            self.restore_checkpoint(checkpoint_path)
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_path={checkpoint_path}`."
            )
        self.predictive.state = CalibStateRepository(checkpoint_dir=checkpoint_path)

    def save_state(
        self, checkpoint_path: Path, keep_top_n_checkpoints: int = 1
    ) -> None:
        return self.predictive.state.put(
            self.predictive.state.get(),
            checkpoint_path=checkpoint_path,
            keep=keep_top_n_checkpoints,
        )

    def _get_output_dim(self, input_shape: Shape, **kwargs) -> int:
        """
        Initialize the state of the joint distribution.

        Parameters
        ----------
        input_shape : Shape
            The shape of the input variable.

        Returns
        -------
        A state of the joint distribution.
        """
        oms = ModelManagerState.init_from_dict(
            self.model_manager.init(input_shape, rng=self.rng.get(), **kwargs)
        )
        inputs = get_inputs_from_shape(input_shape)
        outputs = self.model_manager.apply(oms.params, inputs, mutable=oms.mutable)
        return (
            outputs[0].shape[-1]
            if isinstance(outputs, (list, tuple))
            else outputs.shape[-1]
        )

    def _init(self, data_loader: DataLoader, config: Config):
        for inputs, targets in data_loader:
            input_shape = get_input_shape(inputs)
            break

        state = ModelManagerState.init_from_dict(
            self.likelihood.model_manager.init(input_shape, rng=self.rng.get())
        )
        return CalibState.init(
            params=state.params,
            mutable=state.mutable,
            optimizer=config.optimizer.method,
        )

    def _init_state(self, calib_data_loader: DataLoader, config: Config) -> CalibState:
        if config.checkpointer.restore_checkpoint_path is None:
            if config.checkpointer.start_from_current_state:
                state = self.predictive.state.get(optimizer=config.optimizer.method)
            else:
                state = self._init(calib_data_loader, config)
        else:
            if config.checkpointer.start_from_current_state:
                logging.warning(
                    "`config.checkpointer.start_from_current_state` will be ignored since "
                    "`config.checkpointer.restore_checkpoint_path` is given."
                )
            state = self.restore_checkpoint(
                restore_checkpoint_path=config.checkpointer.restore_checkpoint_path,
                optimizer=config.optimizer.method,
            )
        return state



from fortuna.training.train_state import TrainState


class Callback:
    """
    Base class to define new callback functions. To define a new callback, create a child of this class and
    override the relevant methods.

    Example
    -------
    The following is a custom callback that prints the number of model's parameters at the start of each epoch.

    .. code-block:: python

        class CountParamsCallback(Callback):
            def training_epoch_start(self, state: TrainState) -> TrainState:
                params, unravel = ravel_pytree(state.params)
                logger.info(f"num params: {len(params)}")
                return state
    """

    def training_epoch_start(self, state: TrainState) -> TrainState:
        """
        Called at the beginning of every training epoch

        Parameters
        ----------
        state: TrainState
            The training state

        Returns
        -------
        TrainState
            The (possibly updated) training state
        """
        return state

    def training_epoch_end(self, state: TrainState) -> TrainState:
        """
        Called at the end of every training epoch

        Parameters
        ----------
        state: TrainState
            The training state

        Returns
        -------
        TrainState
            The (possibly updated) training state
        """
        return state

    def training_step_end(self, state: TrainState) -> TrainState:
        """
        Called after every minibatch update

        Parameters
        ----------
        state: TrainState
            The training state

        Returns
        -------
        TrainState
            The (possibly updated) training state
        """
        return state



from typing import List

import jax.numpy as jnp

from fortuna.conformal.regression.base import ConformalRegressor
from fortuna.typing import Array


class CVPlusConformalRegressor(ConformalRegressor):
    """
    This class implements the CV+ method introduced in
    `Barber et al., 2021 <https://www.stat.cmu.edu/~ryantibs/papers/jackknife.pdf>`__. It is an extension of the
    jackknife+ method, introduced in the same work, that consider a K-Fold instead of a leave-one-out strategy. If
    :code:`K=n`, where :code:`n` is the total number of training data, then CV+ reduces to jackknife+.
    """

    def conformal_interval(
        self,
        cross_val_outputs: List[Array],
        cross_val_targets: List[Array],
        cross_test_outputs: List[Array],
        error: float,
    ) -> jnp.ndarray:
        """
        Coverage interval of each of the test inputs, at the desired coverage error. This is supported only for
        one-dimensional target variables.

        Parameters
        ----------
        cross_val_outputs: List[Array]
            Outputs of the models used during cross validation evaluated at their respective validation inputs. More
            precisely, we assume the training data has been jointly partitioned in :code:`K` subsets. The i-th element
            of the list of :code: `cross_val_outputs` is a model trained on all data but the i-th partition, and has
            been evaluated at the inputs of the partition i-th itself, for :code:`i=1, 2, ..., K`.
        cross_val_targets: List[Array]
            Target variables organized in the same partitions used for `cross_val_outputs`. More precisely, the i-th
            element of :code:`cross_val_targets` includes the array of target variables of the i-th partition of the
            training data, for :code:`i=1, 2, ..., K`.
        cross_test_outputs: List[Array]
            Outputs of the models used during cross validation evaluated at the test inputs. More precisely, consider
            the same partition of data as the one used for :code:`cross_val_outputs`. Then the i-th element of
            :code:`cross_test_outputs` represents the outputs of the model that has been trained upon all the training
            data but the i-th partition, and evaluated at the test inputs, for :code:`i=1, 2, ..., K`.
        error: float
            The desired coverage error. This must be a scalar between 0 and 1, extremes included.
        Returns
        -------
        jnp.ndarray
            The conformal intervals. The two components of the second axis correspond to the left and right interval
            bounds.
        """
        if type(cross_val_outputs) != list:
            raise TypeError("`cross_val_outputs` must be a list of arrays.")
        if type(cross_val_targets) != list:
            raise TypeError("`cross_val_targets` must be a list of arrays.")
        if type(cross_test_outputs) != list:
            raise TypeError("`cross_test_outputs` must be a list of arrays.")
        for i, (mu, y, mu_test) in enumerate(
            zip(cross_val_outputs, cross_val_targets, cross_test_outputs)
        ):
            if mu.shape[0] != y.shape[0]:
                raise ValueError(
                    "The first dimension of the i-th element in `cross_val_outputs` must be the same as "
                    "the one of the i-th element in `cross_val_targets`."
                )
            if mu.ndim == 1:
                cross_val_outputs[i] = mu[:, None]
            elif mu.shape[1] != 1:
                raise ValueError(
                    "This method is supported only for scalar model outputs only. However, an element of "
                    "`cross_val_outputs` has second dimension greater than 1."
                )
            if y.ndim == 1:
                cross_val_targets[i] = y[:, None]
            elif y.shape[1] != 1:
                raise ValueError(
                    "This method is supported only for scalar target variables. However, an element of "
                    "`cross_val_targets` has second dimension greater than 1."
                )
            if mu_test.ndim == 1:
                cross_test_outputs[i] = mu_test[:, None]
            elif mu_test.shape[1] != 1:
                raise ValueError(
                    "This method is supported only for scalar model outputs only. However, an element of "
                    "`cross_test_outputs` has second dimension greater than 1."
                )

        r = [jnp.abs(y - mu) for y, mu in zip(cross_val_targets, cross_val_outputs)]
        left = jnp.concatenate(
            [mu[None] - ri[:, None] for mu, ri in zip(cross_test_outputs, r)], 0
        )
        right = jnp.concatenate(
            [mu[None] + ri[:, None] for mu, ri in zip(cross_test_outputs, r)], 0
        )

        qleft = jnp.quantile(left, q=error, axis=0)
        qright = jnp.quantile(right, q=1 - error, axis=0)
        return jnp.array(list(zip(qleft, qright))).squeeze(2)



from typing import (
    Optional,
    Tuple,
)

from flax.core import (
    FrozenDict,
    freeze,
    unfreeze,
)
import flax.linen as nn
from flax.traverse_util import (
    flatten_dict,
    unflatten_dict,
)
import jax
import jax.numpy as jnp
from transformers import DistilBertConfig
from transformers.modeling_flax_utils import ACT2FN
from transformers.models.distilbert.modeling_flax_distilbert import (
    FlaxDistilBertModule,
    FlaxDistilBertPreTrainedModel,
)

from fortuna.model.utils.spectral_norm import WithSpectralNorm


class FlaxSNGPDistilBertExtractorForSequenceClassificationModule(
    nn.Module, WithSpectralNorm
):
    config: DistilBertConfig = None
    dtype: jnp.dtype = jnp.float32

    def setup(self):
        self.distilbert = FlaxDistilBertModule(config=self.config, dtype=self.dtype)
        self.pre_classifier = self.spectral_norm(nn.Dense)(
            self.config.dim,
            dtype=self.dtype,
            kernel_init=jax.nn.initializers.normal(
                stddev=self.config.initializer_range
            ),
        )
        self.dropout = nn.Dropout(rate=self.config.seq_classif_dropout)

    def __call__(
        self,
        input_ids,
        attention_mask,
        deterministic: bool = True,
        output_attentions: bool = False,
        output_hidden_states: bool = False,
        return_dict: bool = True,
    ):
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        # Model
        distilbert_output = self.distilbert(
            input_ids,
            attention_mask,
            deterministic=deterministic,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_state = distilbert_output[0]  # (bs, seq_len, dim)
        pooled_output = hidden_state[:, 0]  # (bs, dim)
        pooled_output = self.pre_classifier(pooled_output)  # (bs, dim)
        pooled_output = ACT2FN["relu"](pooled_output)
        pooled_output = self.dropout(pooled_output, deterministic=deterministic)

        if not return_dict:
            return (pooled_output,) + distilbert_output[1:]

        return pooled_output


class FlaxSNGPDistilBertExtractorForSequenceClassification(
    FlaxDistilBertPreTrainedModel
):
    module_class = FlaxSNGPDistilBertExtractorForSequenceClassificationModule

    def init_weights(
        self, rng: jax.random.PRNGKey, input_shape: Tuple, params: FrozenDict = None
    ) -> FrozenDict:
        # init input tensors
        input_ids = jnp.zeros(input_shape, dtype="i4")
        attention_mask = jnp.ones_like(input_ids)

        params_rng, dropout_rng = jax.random.split(rng)
        rngs = {"params": params_rng, "dropout": dropout_rng}

        random_params = self.module.init(
            rngs, input_ids, attention_mask, return_dict=False
        )
        random_params = FrozenDict(
            {
                "spectral_stats": random_params["spectral_stats"].unfreeze(),
                **random_params["params"].unfreeze(),
            }
        )
        if params is not None:
            random_params = flatten_dict(unfreeze(random_params))
            params = flatten_dict(unfreeze(params))
            for missing_key in self._missing_keys:
                params[missing_key] = random_params[missing_key]
            self._missing_keys = set()
            return freeze(unflatten_dict(params))
        else:
            return random_params

    def __call__(
        self,
        input_ids,
        attention_mask=None,
        head_mask=None,
        params: dict = None,
        mutable: dict = None,
        dropout_rng: jax.random.PRNGKey = None,
        train: bool = False,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.return_dict
        )

        if attention_mask is None:
            attention_mask = jnp.ones_like(input_ids)

        # Handle any PRNG if needed
        rngs = {}
        if dropout_rng is not None:
            rngs["dropout"] = dropout_rng

        variables = {"params": params} if params is not None else self.params
        if mutable:
            mutable_variables = mutable["model"].unfreeze()
            variables.update(mutable_variables)
            mutable = list(mutable_variables.keys())

        return self.module.apply(
            variables,
            jnp.array(input_ids, dtype="i4"),
            jnp.array(attention_mask, dtype="i4"),
            not train,
            output_attentions,
            output_hidden_states,
            return_dict,
            mutable=mutable if train else False,
            rngs=rngs if train else None,
        )
